/*  strassen.c   by C.W.Kessler 06/99 */

#include <fork.h>
#include <syscall.h>
#include <io.h>
#include <assert.h>
#include <Matrix.h>

extern sync void StrassenMul( sh Matrix A, sh Matrix B, sh Matrix C );

#define GROUPHEAP 0
#define GLOBALHEAP 0

/* comp0(), ..., comp6()  implement Strassen's set of formulae: */

sync void comp0( sh Matrix q0, sh Matrix **a, sh Matrix **b )
{
  sh int ndiv2 = a[0][0]->nrows;
  sh Matrix t1 = new_Matrix( ndiv2, ndiv2, GLOBALHEAP );
  sh Matrix t2 = new_Matrix( ndiv2, ndiv2, GLOBALHEAP );
  Add( a[0][0], a[1][1], t1 );
  Add( b[0][0], b[1][1], t2 );
  StrassenMul( t1, t2, q0 );
  seq { freeMatrix(t1); freeMatrix(t2); }
}


sync void comp1( sh Matrix q1, sh Matrix **a, sh Matrix **b )
{
  sh int ndiv2 = a[0][0]->nrows;
  sh Matrix t1 = new_Matrix( ndiv2, ndiv2, GLOBALHEAP );
  Add( a[1][0], a[1][1], t1 );
  StrassenMul( t1, b[0][0], q1 );
  seq freeMatrix(t1); 
}
  

sync void comp2( sh Matrix q2, sh Matrix **a, sh Matrix **b )
{
  sh int ndiv2 = a[0][0]->nrows;
  sh Matrix t1 = new_Matrix( ndiv2, ndiv2, GLOBALHEAP );
  Sub( b[0][1], b[1][1], t1 );
  StrassenMul( a[0][0], t1, q2 );
  seq freeMatrix(t1); 
}


sync void comp3( sh Matrix q3, sh Matrix **a, sh Matrix **b )
{
  sh int ndiv2 = a[0][0]->nrows;
  sh Matrix t1 = new_Matrix( ndiv2, ndiv2, GLOBALHEAP );
  Sub( b[1][0], b[0][0], t1 );
  StrassenMul( a[1][1], t1, q3 );
  seq freeMatrix(t1); 
}


sync void comp4( sh Matrix q4, sh Matrix **a, sh Matrix **b )
{
  sh int ndiv2 = a[0][0]->nrows;
  sh Matrix t1 = new_Matrix( ndiv2, ndiv2, GLOBALHEAP );
  Add( a[0][0], a[0][1], t1 );
  StrassenMul( t1, b[1][1], q4 );
  seq freeMatrix(t1); 
}
  

sync void comp5( sh Matrix q5, sh Matrix **a, sh Matrix **b )
{
  sh int ndiv2 = a[0][0]->nrows;
  sh Matrix t1 = new_Matrix( ndiv2, ndiv2, GLOBALHEAP );
  sh Matrix t2 = new_Matrix( ndiv2, ndiv2, GLOBALHEAP );
  Sub( a[1][0], a[0][0], t1 );
  Add( b[0][0], b[0][1], t2 );
  StrassenMul( t1, t2, q5 );
  seq { freeMatrix(t1); freeMatrix(t2); }
}


sync void comp6( sh Matrix q6, sh Matrix **a, sh Matrix **b )
{
  sh int ndiv2 = a[0][0]->nrows;
  sh Matrix t1 = new_Matrix( ndiv2, ndiv2, GLOBALHEAP );
  sh Matrix t2 = new_Matrix( ndiv2, ndiv2, GLOBALHEAP );
  Sub( a[0][1], a[1][1], t1 );
  Add( b[1][0], b[1][1], t2 );
  StrassenMul( t1, t2, q6 );
  seq { freeMatrix(t1); freeMatrix(t2); }
}


sync int issmall( sh int n )
{
  if (n <= 2) return 1;
  else        return 0;
}

sync void comp( sh int i, sh Matrix *q, sh Matrix **a, sh Matrix **b )
{
 //farm pprintf("call comp%d %dX%d\n", i, q[i]->nrows, q[i]->ncols );
 if (i==0)      comp0( q[i], a, b );
 else if (i==1) comp1( q[i], a, b );
 else if (i==2) comp2( q[i], a, b );
 else if (i==3) comp3( q[i], a, b );
 else if (i==4) comp4( q[i], a, b );
 else if (i==5) comp5( q[i], a, b );
 else           comp6( q[i], a, b );
}

sync void StrassenMul( sh Matrix A, sh Matrix B, sh Matrix C )
{
 // Works only for square matrices A, B where n=m contains a 
 // sufficiently large power of 2. Matrix C is pre-allocated.
 sh int n = A->nrows,  ndiv2 = n / 2;
 sh Matrix **a, **b, **c, t[2][2], q[7]; 
 sh int issm,  p = #; 
 int i;
 
 if (p==1) { Mul( A, B, C ); return; } //not enough processors
                             //available for a recursion step
 issm = issmall(n);
 if (issm) { Mul( A, B, C ); return; }    // the trivial case
 
 // create 4 quarter-submatrix views of each matrix:
 a = splitMatrix( A, ndiv2, ndiv2 ); 
 b = splitMatrix( B, ndiv2, ndiv2 );
 c = splitMatrix( C, ndiv2, ndiv2 );

 if (p >= 7) 
   fork ( 7; @ = $$%7; ) {  // subgroup @ computes q@: 
     //farm pprintf("Case 1: ndiv2 = %d\n", ndiv2);
     q[@] = new_Matrix( ndiv2, ndiv2, GLOBALHEAP );
     comp( @, q, a, b );
   }
 else 
   fork ( p; @=$$; ) {
     sh int j;
     for( j=@; j<7; j+=p ) {
       //farm pprintf("Case 2: ndiv2 = %d\n", ndiv2);
       q[j] = new_Matrix( ndiv2, ndiv2, GLOBALHEAP );
       comp( j, q, a, b );
     }
   }
 if (p>=4)
   fork( 4; @= $$%4; ) {
      //farm pprintf("Case 3: ndiv2 = %d\n", ndiv2);
      t[@/2][@%2] = new_Matrix( ndiv2, ndiv2, GLOBALHEAP );
   }
 else {
   sh int j;
   for( j=0; j<4; j++ ) {
      //farm pprintf("Case 4: ndiv2 = %d\n", ndiv2);
      t[j/2][j%2] = new_Matrix( ndiv2, ndiv2, GLOBALHEAP );
   }
 }
 if ($$ % 2) {
      Add( q[0], q[3], t[0][0] );
      Sub( q[6], q[4], t[0][1] );
      Add( t[0][0], t[0][1], c[0][0] );
      Add( q[1], q[3], c[1][0] );
 }
 else {               Add( q[2], q[4], c[0][1] );
                      Add( q[0], q[2], t[1][0] );
                      Sub( q[5], q[1], t[1][1] );
                      Add( t[1][0], t[1][1], c[1][1] );
 }
 farm {
   forall(i,0,7,p) freeMatrix( q[i] );
   forall(i,0,4,p) freeMatrix( t[i/2][i%2] );
 }
}
 

void main( void ) 
{
 start {  // if ($<343)       /* this starts 7^3 processors */
   sh int N = 8;  // must be a power of 2 or at least contain 2^(log_7 p)
   sh Matrix A, B, C;
   sh int p = groupsize();
   sh int t;
   int i, j;
  
   seq {
      prS("STRASSEN MATRIX-MATRIX MULTIPLICATION\n\n");
      printf("Enter matrix size N: ");
      scanf("%d", &N);
   }
   A = new_Matrix( N, N, GLOBALHEAP );
   B = new_Matrix( N, N, GLOBALHEAP );
   C = new_Matrix( N, N, GLOBALHEAP );
   /* preset the input array: */
   farm
      forall( i, 0, N, p )
         for (j=0; j<N; j++) { 
            Set(A, i,j, 1.0);
            Set(B, i,j, 2.0);
         }
   seq {
      prS("\nMatrix A:\n");
      printMatrix( A );
      prS("\nMatrix B:\n");
      printMatrix( B );
      t = getct();
   }
   StrassenMul( A, B, C );
   seq {
      t = getct() - t;
      prS("\nMatrix Product:\n");
      printMatrix( C );
      printf("\nTime: %d msecs\n", t>>8 );
      t = getct();
   } 
   Mul( A, B, C );
   seq {
      t = getct() - t;
      prS("\nStandard Matrix Product:\n");
      printMatrix( C );
      printf("\nO(n^3) Time: %d msecs\n<press key to continue>", t>>8 );
      scanf("%c", &t);
   } 
 }
}

