// mpi_trv: TRACEFILE VISUALIZER FOR FORK95 TRACEFILES
// including FORK_MPI support
// v0013 Christoph W. Kessler, Oliver Fritzen
// 07/04/00
// 

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#define min(a,b) ((a)<=(b)?(a):(b))
#define max(a,b) ((a)>=(b)?(a):(b))

// input: a Fork95 tracefile generated by writeTraceFile("name","title")
// output: an xfig source file for a graphical visualization of the input

// These defines must be the same as in /include/fork.h:




enum {
#define TRACE_ARROW(A,B,C,D)
#define COLOR_RANGE(A,B,C,D,E,F,G,H,I)
#define TRACE_EVENT(A,B,C,D,E,F,G) B=A,
#include "trace.h"
};

struct {   /* global parameters */
  float x_scale;
  double ratio;
  int bar_width;
  int nostats;
  char *conf_file_name;
  char *infile;
  char *outfile;
} glob_args;

struct {   /* event properties */
  int color;
  int fill_color;
  float y_size;
  float y_offset;
  int pattern;
} event_props[TRACE_MAX_EVENT];

struct {   /* arrow properties */
  int inner_width;
  int inner_color;
  int outer_width;
  int outer_color;
} arrow_props;

typedef struct {  /* color table */
  int r1;
  int r2;
  int g1;
  int g2;
  int b1;
  int b2;
  int s1;
  int s2;
  float skew_power;
} color_table_type;

color_table_type color_table;

extern char yytext[]; // needed for flex input

int shms;
// COLORED defined in Makefile   // default: greyscale diagram
#define MHz 8.0        /* SBPRAM prototype */
#define vPs 32.0
#define NCOLORS 192   // user-defined colors in xfig color table
#define NGREYSCALES 12 // grey scales in xfig format



writePrologue( FILE *fp )
{
  fprintf(fp,"#FIG 3.2\nLandscape\nCenter\nInches\nLetter\n100.00\nSingle\n-2\n1200 2\n");
}

writeColorTable( FILE *fp)
{
  int i;
  color_table_type *c;
  c = &color_table;
  for (i=0; i<NCOLORS; i++) {
    fprintf(fp,"0 %d #%2.2x%2.2x%2.2x\n", i+32, 
	    color_table.r1 + ((color_table.r2 - color_table.r1) *i)/NCOLORS,
	    color_table.g1 + ((color_table.g2 - color_table.g1) *i)/NCOLORS,
	    color_table.b1 + ((color_table.b2 - color_table.b1) *i)/NCOLORS);
  }
}


void die(const char *str) {
  fprintf(stderr, str);
  exit(1);
}

int parse_config_file(const char *filename) {
  FILE *conf_file;
  int i,nr, tok_nr = 1, tok_state = 1;
  char namestring[80];
  
  if(!(conf_file = fopen(filename,"r"))) return 1;
  yyrestart(conf_file);
  while(tok_nr = yylex()) {
    if(tok_nr != 8) {
      //fprintf(stderr, "state = %d, token = '%s', token_nr = %d\n", tok_state, yytext, tok_nr);
      switch(tok_state) {
      case 1:
	if(tok_nr == 3) tok_state = 3; // TRACE_EVENT
	else if(tok_nr == 4) tok_state = 10; // COLOR_RANGE
	else if(tok_nr == 5) tok_state = 2; // begin of comment
	else if(tok_nr == 9) tok_state = 19; // TRACE_ARROW
	else die("Wrong input file format, exiting...\n");
	break;
      case 2: 
	if(tok_nr == 6) tok_state = 1; // end of comment  
	break;
      case 3:
	if(tok_nr == 1) {
	  nr =  atoi(yytext);
	  tok_state++;
	  break;
	}
	else die("Wrong input file format, exiting...\n" );
      case 4:
	if(tok_nr == 7) {
	  strcpy(namestring, yytext);
	  tok_state++;
	  break;
	}
	else die("Wrong input file format, exiting...\n" );
      case 5:
	if(tok_nr == 1) {
	  event_props[nr].color =  atoi(yytext);
	  tok_state++;
	  break;
	}
	else die("Wrong input file format, exiting...\n" );
      case 6:
	if(tok_nr == 1) {
	  event_props[nr].fill_color =  atoi(yytext);
	  tok_state++;
	  break;
	}
	else die("Wrong input file format, exiting...\n" );
      case 7:
	if(tok_nr == 2) {
	  event_props[nr].y_size =  atof(yytext);
	  tok_state++;
	  break;
	}
	else die("Wrong input file format, exiting...\n" );
      case 8:
	if(tok_nr == 2) {
	  event_props[nr].y_offset =  atof(yytext);
	  tok_state++;
	  break;
	}
	else die("Wrong input file format, exiting...\n" );
      case 9:
	if(tok_nr == 1) {
	  event_props[nr].pattern =  atoi(yytext);
	  tok_state = 1;
	  break;
	}
	else die("Wrong input file format, exiting...\n" );
      case 10:
	if(tok_nr == 1) {
	  color_table.r1 =  atoi(yytext);
	  tok_state++;
	  break;
	}
	else die("Wrong input file format, exiting...\n" );
      case 11:
	if(tok_nr == 1) {
	  color_table.r2 =  atoi(yytext);
	  tok_state++;
	  break;
	}
	else die("Wrong input file format, exiting...\n" );
      case 12:
	if(tok_nr == 1) {
	  color_table.g1 =  atoi(yytext);
	  tok_state++;
	  break;
	}
	else die("Wrong input file format, exiting...\n" );
      case 13:
	if(tok_nr == 1) {
	  color_table.g2 =  atoi(yytext);
	  tok_state++;
	  break;
	}
	else die("Wrong input file format, exiting...\n" );
      case 14:
	if(tok_nr == 1) {
	  color_table.b1 =  atoi(yytext);
	  tok_state++;
	  break;	
	}
	else die("Wrong input file format, exiting...\n" );
      case 15:
	if(tok_nr == 1) {
	  color_table.b2 =  atoi(yytext);
	  tok_state++;
	  break;
	}
	else die("Wrong input file format, exiting...\n" );
      case 16:
	if(tok_nr == 1) {
	  color_table.s1 =  atoi(yytext);
	  tok_state++;
	  break;
	}
	else die("Wrong input file format, exiting...\n" );
      case 17:
	if(tok_nr == 1) {
	  color_table.s2 =  atoi(yytext);
	  tok_state++;
	  break;
	}
	else die("Wrong input file format, exiting...\n" );
      case 18:
	if(tok_nr == 2) {
	  color_table.skew_power =  atof(yytext);
	  tok_state = 1;
	  break;
	}
	else die("Wrong input file format, exiting...\n" );
      case 19:
	if(tok_nr == 1) {
	  arrow_props.inner_width = atoi(yytext);
	  tok_state++;
	  break;
	}
	else die("Wrong input file format, exiting...\n" );
      case 20:
	if(tok_nr == 1) {
	  arrow_props.inner_color = atoi(yytext);
	  tok_state++;
	  break;
	}
	else die("Wrong input file format, exiting...\n" );
      case 21:
	if(tok_nr == 1) {
	  arrow_props.outer_width = atoi(yytext);
	  tok_state++;
	  break;
	}
	else die("Wrong input file format, exiting...\n" );
      case 22:
	if(tok_nr == 1) {
	  arrow_props.outer_color = atoi(yytext);
	  tok_state = 1;
	  break;
	}
	else die("Wrong input file format, exiting...\n" );
      }
    }
  }
  fclose(conf_file);
  return 0;
}


inline int color( g ) {
  int c;
  float ratio = 1.0 - (double)g/(double)shms;
  ratio = 1.0 - pow(ratio, (double) (color_table.skew_power)); 
  // skew range towards green
  if (g >= shms) //  heap location in joins:
    return 32+NCOLORS-1;  // g = 2*shms - g;
  c = 32 + (int)((float)NCOLORS*ratio);
  if (c<32 || c>32+NCOLORS)
    fprintf(stderr, "Error: unknown color %d\n", c );
  return c;
}

void doargs(argc,argv)
  int argc;
char *argv[];
{
  char *s;
  char *string1;
  char *string2;
  int i;
  
  string2 = NULL;
  glob_args.x_scale = 1.0;
  glob_args.ratio = 1.5;
  glob_args.bar_width = 240;
  glob_args.nostats = 0;
  glob_args.conf_file_name = NULL;
  glob_args.infile = NULL;
  glob_args.outfile = NULL;
  for( i=1; i<argc; i++ ) {
    s = argv[i];
    if (s[0]=='-') {  /* Option */
      string1 = string2 = NULL;
      switch( s[1] ) {
      case 'x':case 'X':
	if( i+1 < argc ) glob_args.x_scale = atof( argv[++i] );
	//fprintf( stderr, "x_scale = %d\n", glob_args.x_scale);
	break;
      case 'r':case 'R':
	if( i+1 < argc ) glob_args.ratio = atof( argv[++i] );
	break;
      case 'w':case 'W':
	if( i+1 < argc ) glob_args.bar_width = atoi( argv[++i] );
	break;
      case 'c':case 'C':
	if( i+1 < argc ) {
	  glob_args.conf_file_name = 
	    (char *) malloc( strlen( argv[++i] ) + 1 );
	  strcpy( glob_args.conf_file_name, argv[i] );
	}
	break;
      case 'n':case 'N':
	glob_args.nostats = 1;
	break;
      default:
	die("usage: trv [-x xsize] [-r ratio] [-w barwidth] [-c config_file] [-nostats] [input file] [output file]\n");
      }
    }
    else {
      string1 = string2;
      string2 = s;
    }
  }
  if(string2 == NULL) {
    glob_args.infile = (char *) malloc( 10 );  
    strcpy( glob_args.infile, "trace.trv" );
    glob_args.outfile = (char *) malloc( 10 );  
    strcpy( glob_args.outfile, "trace.fig" );
  }
  else if(string1 == NULL) {
    glob_args.infile = (char *) malloc( strlen( string2 ) + 1 );  
    strcpy( glob_args.infile, string2 );
    glob_args.outfile = (char *) malloc( strlen( string2 ) + 1 );  
    strncpy( glob_args.outfile, string2, strlen( string2 ) - 3);
    strcpy( ( glob_args.outfile + strlen( string2 ) - 3), "fig");
  }
  else {
    glob_args.infile = (char *) malloc( strlen( string1 ) + 1 );  
    strcpy( glob_args.infile, string1 );
    glob_args.outfile = (char *) malloc( strlen( string2 ) + 1 );  
    strcpy( glob_args.outfile, string2 );
  }
}


int main ( int argc, char **argv ) {
  FILE *fp, *out;
  char *title, *filename, *outfilename,*path;
  int nprocs, t0, tf, firstgps;
  int shloads, shstores, shmpadds, shmpmaxs, shmpands, shmpors;
  int *shload, *shstore, *mpadd, *mpmax, *mpand, *mpor;  // local counters
  int *sendscope, *recvscope, *lastArrowStartID, *lastevent_offs, 
    *lastevent_color, *lastevent_fill_color, *lastevent_pattern,
    *lastevent_saturation;
  
  int *nbarriers, // number of barriers executed in trace period
    *tbarriers; // accumulated time each proc. spends on barrier wait
  int *lastbarriertime; // start of last barrier event of each processor
  int *nlockups, // number of lockups executed in trace period
    *tlockups; // accumulated time each proc. spends on lockup wait
  int *lastlockuptime; // start of last lockup event of each processor
  int xmax = 13000, xoff = 600,
    ymax,         yoff = 550;//alt: 450;
  int rborder = 50, bborder = 50;
  int dropP = 40; // Pi moved downwards from graphics line y coordinate
  int ycorr = 25; // move upwards the inbetween-text by that many points
  int tdist = xmax - xoff - rborder;   // horizontal space for bars
  int inbetw = 180;     // vertical space between processor bars
  int width;      // width of a processor bar 
  int evpid, ev, pid, t, gps, i;
  int xli, xre, yli, yre;
  char *dotpos; // 990817
  float ftdist;

#define Xoffset( tt ) ((int)(tdist*(float)((tt)-t0) * glob_args.x_scale/ ftdist) + xoff )
#define Yoffset( pp ) ((pp)*(width+inbetw) + yoff)
  
  doargs(argc, argv);
  if(parse_config_file("trace.h")) {
    path = strcat(getenv("FORKDIR"),"/lib/trace.h");
    if(parse_config_file(path)) 
      die(strcat("Cannot open/cannot find file ",path));
  }
  if(glob_args.conf_file_name) 
    parse_config_file(glob_args.conf_file_name);
  filename = glob_args.infile;
  outfilename = glob_args.outfile;
  if(! (fp = fopen(filename,"r"))) {
    fprintf( stderr, "Cannot open input file %s\n", filename);
    exit(1);
  }
  if(! (out = fopen(outfilename,"w"))) {
    fprintf( stderr, "Cannot open output file %s\n", outfilename);
    exit(1);
  }
  fscanf( fp, "%d %x %x %x %x %d %d %d %d %d %d \n",
	  &nprocs, &t0, &tf, &shms, &firstgps, 
	  &shloads, &shstores, &shmpadds, &shmpmaxs, &shmpands, &shmpors );
  // and forget these again, read them only for backward compatibility
  // version 2.0 will remove the entries shloads..shmpors,
  // as these are computed directly from the local contributions
  
  xmax = (int) ((double) (xmax-xoff) * glob_args.x_scale) + xoff;
  ymax = (int) ((double) xmax * glob_args.ratio + yoff); 
  width = max(30, min(glob_args.bar_width,
	      (int)((double)(ymax-yoff)/(double)nprocs) - inbetw));
  
  //fprintf(stderr,"width = %d, inbetw = %d\n", width, inbetw);
  
  title = (char *)malloc(100);
  if (!fgets( title, 100, fp ))
    printf("\ntrv: warning: error in reading title\n");
  printf("\ntrv: creating xfig source %s from tracefile %s\n", 
	   outfilename, filename );
  
  printf("nprocs=%d starttime=%d stoptime=%d shmemsize=%d firstgps=%d title=%s\n",
         nprocs, t0, tf, shms, firstgps, title );
  if (nprocs <= 0) { printf("trv: error: wrong number of processors\n"); exit(1); }
  if (shms<=1) { printf("trv: error: wrong shared memory size\n"); exit(1); }
  if (t0>=tf) { printf("trv: error: starttime must precede stoptime\n"); exit(1); }
  ftdist = (float)(tf - t0);
  
  writePrologue( out );
  writeColorTable( out );
  
  // draw black(7) rectangle as bounding box:
  /* fprintf(out, */
  /* 	  "2 2 0 1 0 7 0 0 -1 0.000 0 0 -1 0 0 5\n\ */
  /*               0 0 %d 0 %d %d 0 %d 0 0\n", xmax, xmax,  */
  /* 	  Yoffset(nprocs) + yoff, Yoffset(nprocs) + yoff ); */
  
  // read statistics:
  shstore = (int *)malloc( nprocs * sizeof(int));
  shload  = (int *)malloc( nprocs * sizeof(int));
  mpadd = (int *)malloc( nprocs * sizeof(int));
  mpmax = (int *)malloc( nprocs * sizeof(int));
  mpand = (int *)malloc( nprocs * sizeof(int));
  mpor = (int *)malloc( nprocs * sizeof(int));
  shloads = shstores = shmpadds = shmpmaxs = shmpands = shmpors = 0;
  for (i=0; i<nprocs; i++) {
    fscanf( fp, "%x %x %x %x %x %x\n",
            shload+i, shstore + i, mpadd+i, mpmax+i, mpand+i, mpor+i );
    shloads += shload[i];
    shstores += shstore[i];
    shmpadds += mpadd[i];
    shmpmaxs += mpmax[i];
    shmpands += mpand[i];
    shmpors += mpor[i];
  }
  // initialize lastevent_color lastevent_fill_color lastevent_offs lastevent_pattern:
  lastArrowStartID = (int *)malloc( nprocs * sizeof(int) );
  sendscope = (int *)malloc( nprocs * sizeof(int) );
  recvscope = (int *)malloc( nprocs * sizeof(int) );
  for (i=0; i<nprocs; i++) {
    lastArrowStartID[i] = 0;
    sendscope[i]=0;
    recvscope[i]=0;
  }
  lastevent_offs = (int *)malloc( nprocs * sizeof(int) );
  lastevent_color = (int *)malloc( nprocs * sizeof(int) );
  lastevent_fill_color = (int *)malloc( nprocs * sizeof(int) );
  lastevent_pattern = (int *)malloc( nprocs * sizeof(int) );
  lastevent_saturation = (int *)malloc( nprocs * sizeof(int) );
  
  for (i=0; i<nprocs; i++) {
    lastevent_offs[i] = Xoffset(t0);
    lastevent_color[i] = color( firstgps );
    lastevent_fill_color[i] = color( firstgps );
    lastevent_pattern[i] = 20;
    lastevent_saturation[i] = 50;
  }
  nbarriers = (int *)malloc( nprocs * sizeof(int) );
  tbarriers = (int *)malloc( nprocs * sizeof(int) );
  lastbarriertime = (int *)malloc( nprocs * sizeof(int) );
  nlockups = (int *)malloc( nprocs * sizeof(int) );
  tlockups = (int *)malloc( nprocs * sizeof(int) );
  lastlockuptime = (int *)malloc( nprocs * sizeof(int) );
  for (i=0; i<nprocs; i++) {
    nbarriers[i] = tbarriers[i] = lastbarriertime[i] =
      nlockups[i] = tlockups[i] = lastlockuptime[i] = 0;
  }
  printf("reading events...\n");
  while (!feof(fp)) {
    int k = fscanf( fp, "%x %x %x\n", &evpid, &t, &gps );
    if (k<3) break;
    ev = evpid >> 16;     // event type
    if (ev==TRACE_END) break;   // end marker found
    if (ev<0 || ev>31) {
      printf("Error: wrong event ID: %d (was: %d)\n", ev, evpid );
      exit(1);
    }
    pid = evpid & 0xffff; // processor ID
    if (pid >= nprocs || pid < 0 ) {
      printf("Error: wrong processor ID: %d\n", pid );
      exit(1);
    }
    xli = lastevent_offs[pid];      
    xre = Xoffset( t );
    if(recvscope[pid]) {
      yli = Yoffset(pid) + (width>>1);
      yre = yli + (width>>1);
    } 
    else if(sendscope[pid]) {
      yli = Yoffset(pid);
      yre = yli + (width>>1);
    }
    else {
      yli = Yoffset(pid);
      yre = yli + width;
    }
    /* Begin drawing arrow */
    if(ev==TRACE_ENTER_SEND) sendscope[pid] = 1;
    if(ev==TRACE_EXIT_SEND) sendscope[pid] = 0;
    if(ev==TRACE_ENTER_RECV) {
      recvscope[pid] = 1;
      lastArrowStartID[pid] = gps;      
    }
    if(ev==TRACE_EXIT_RECV) {
      recvscope[pid] = 0;
      if(arrow_props.outer_width > 0) 
	fprintf( out,
		"2 1 0 %d %d 7 0 0 20 0.000 0 0 -1 1 0 2\n
0 0 1.00 60.00 120.00\n
%d %d %d %d\n",arrow_props.outer_width, arrow_props.outer_color, 
		Xoffset(gps), Yoffset(lastArrowStartID[pid]) + (width>>1),
		Xoffset(t), Yoffset(pid)+ (width>>1));
      fprintf( out,
	      "2 1 0 %d %d 7 0 0 20 0.000 0 0 -1 1 0 2\n
0 0 1.00 60.00 120.00\n
%d %d %d %d\n",
	      arrow_props.inner_width, arrow_props.inner_color,
	      Xoffset(gps), Yoffset(lastArrowStartID[pid]) + (width>>1),
	      Xoffset(t), Yoffset(pid)+ (width>>1));
    } /* End drawing arrow */
    fprintf( out,
	     "2 2 0 1 %d %d %d 0 %d 0.000 0 0 -1 0 0 5\n\
%d %d %d %d %d %d %d %d %d %d\n",
	     lastevent_color[pid],
	     lastevent_fill_color[pid], 
	     lastevent_saturation[pid],
	     lastevent_pattern[pid],
	     xli, yli, xre, yli, xre, yre, xli, yre, xli, yli );
    lastevent_offs[pid] = xre;
    
    if (ev==TRACE_BARRIER_ENTRY) {
      lastbarriertime[pid] = t;
      nbarriers[pid] ++;
    }
    if (ev==TRACE_BARRIER_EXIT)
      if (lastbarriertime[pid] > 0) { 
	tbarriers[pid] += (t - lastbarriertime[pid]);
      }
    if (ev==TRACE_LOCKUP_WAIT) {
      lastlockuptime[pid] = t;
      nlockups[pid] ++;
    }
    if (ev==TRACE_LOCKUP_DONE)
      if (lastlockuptime[pid] > 0) { // ignore the exitbarrier of the leader
	// processor that executed startTracing
	tlockups[pid] += (t - lastlockuptime[pid]);
      }
    if (ev==TRACE_GROUP || ev==TRACE_BARRIER_EXIT || ev==TRACE_LOCKUP_DONE) {
      lastevent_color[pid] = color( gps );
      lastevent_fill_color[pid] = color( gps );
    }
    else if((!(ev==TRACE_ENTER_SEND)) &&
	    (!(ev==TRACE_ENTER_RECV)) &&
	    (!(ev==TRACE_EXIT_SEND)) &&
	    (!(ev==TRACE_EXIT_RECV))) {
      //fprintf(stderr,"Warning: unknown eventtype %d\n",ev);
      //lastevent_color[pid] = ev;   // user-defined in 8..31
    }
    //debug: printf("Event: %d %d %d %d => x=%d\n", ev, pid, t, gps, xre );
    if(event_props[ev].color != -1) 
      lastevent_color[pid] = event_props[ev].color;
    if(event_props[ev].fill_color != -1) 
      lastevent_fill_color[pid] = event_props[ev].fill_color;
    if(event_props[ev].pattern != -1) 
      lastevent_pattern[pid] = event_props[ev].pattern;
    if(lastevent_fill_color[pid] >= 32) 
      lastevent_saturation[pid] = 
	color_table.s1 + 
	(int)
	(((float)(color_table.s2 - color_table.s1)/(float)NCOLORS ) * 
	 (float)lastevent_fill_color[pid]);
    else lastevent_saturation[pid] = 50;
  }
  fclose(fp);
  // finalize for each processor:
  for (i=0; i<nprocs; i++) {
    xli = lastevent_offs[i];      xre = Xoffset( tf );
    yli = Yoffset(i);             yre = yli + width;
    
    fprintf( out, "2 2 0 1 %d %d 0 0 %d 0.000 0 0 -1 0 0 5\n\
%d %d %d %d %d %d %d %d %d %d\n",
	     lastevent_color[i], lastevent_fill_color[i],
	     lastevent_pattern[i], 
	     xli, yli, xre, yli, xre, yre, xli, yre, xli, yli );
  }
  // Logo box:
  fprintf(out,"2 2 0 1 31 31 4 0 20 0.000 0 0 -1 0 0 5\n");
  fprintf(out,"         45 45 550 45 550 440 45 440 45 45\n");
  
  // draw title
  fprintf(out,"4 0 0 0 0 18 20 0.0000 4 225 825 630 460 %s\\001\n", title );
  fprintf(out,"4 2 0 0 0 22 14 0.0000 4 225 825 %d 440 traced time period: %d msecs\\001\n",
	  xmax-50,
	  (int)(ftdist / MHz / 1000.0 * vPs) );
  fprintf(out,"4 0 0 0 0 20 12 0.0000 4 225 825 630 200 %d sh-loads,  %d sh-stores,  %d mpadd,  %d mpmax,  %d mpand,  %d mpor\\001\n", 
	  shloads, shstores, shmpadds, shmpmaxs, shmpands, shmpors);
  // draw processors:
  for (i=0; i<nprocs; i++)
    fprintf(out,"4 0 0 0 0 22 %d 0.0000 4 100 440 75 %d P%d\\001\n",
	    14, 
	    Yoffset(i) + width + dropP, i );
  for (i=0; i<nprocs; i++) {
    // draw barrier statistics:
    fprintf(out,"4 0 0 0 0 20 %d 0.0000 4 100 440 %d %d %d barriers, \
%2.1f%% spent spinning on barriers\\001\n",
	    10,
	    xoff, Yoffset(i) + width + inbetw - ycorr,
	    nbarriers[i], //(int)( (float)(tbarriers[i])/MHz/1000.0*vPs),
	    (float)(tbarriers[i]) / ftdist * 100.0 );
    // draw lockup statistics:
    fprintf(out,"4 0 0 0 0 20 %d 0.0000 4 100 440 %d %d %d lockups, \
%2.1f%% spent spinning on locks\\001\n",
	    10,
	    xoff+3600, Yoffset(i) + width + inbetw - ycorr,
	    nlockups[i], //(int)( (float)(tlockups[i])/MHz/1000.0*vPs),
	    (float)(tlockups[i]) / ftdist * 100.0 );
    if(! glob_args.nostats) {
      // draw shared memory access statistics:
      fprintf(out,"4 2 0 0 0 20 %d 0.0000 4 100 440 %d %d %d sh loads, \
%d sh stores,   %d mpadd,  %d mpmax,  %d mpand,  %d mpor\\001\n",
	      10,
	      xmax-50, Yoffset(i) + width + inbetw - ycorr, 
	      shload[i], shstore[i], mpadd[i], mpmax[i], mpand[i], mpor[i] );
    }
  }
  // Logobeschriftung:
    fprintf(out,"4 0  8 0 0 22 12 0.0000 4 135 495 114 195 Fork\\001\n");
    fprintf(out,"4 0 18 0 0 19 18 0.0000 4 180 345 114 415 trv\\001\n");
  fclose(out);
  return 0;
}
  
