/** Matrix.c
 *
 *  basic matrix functions on Matrix datatype
 *
 *  C. Kessler reworked 991202
 */
#include <fork.h>   // always required in Fork programs
#include <io.h>     // printf() prototype 
#include <string.h> // memcpy() prototype
#include <assert.h>
#include <Matrix.h>

#define GROUPHEAP 1

sync Matrix new_Matrix( sh int nr, sh int nc, sh int theheap )
{
 sh Matrix M;
 sh float *buf;
 int i;
 
 if (theheap == GROUPHEAP) {
   seq {
     assert( nr > 0 );
     assert( nc > 0 );
   }
   M = (Matrix)shalloc( sizeof(matrix) );
   M->row = (float **)shalloc( nr * sizeof(float *) );
   M->data = buf = (float *)shalloc( nr * nc * sizeof(float) );
 }
 else
  seq {
   assert( nr > 0 );
   assert( nc > 0 );
   //if (nr<=0 || nc<=0) pprintf("ERROR! new_Matrix(%d,%d,%d)\n", nr, nc, theheap);
   M = (Matrix)shmalloc( sizeof(matrix) );
   assert(M);
   M->row = (float **)shmalloc( nr * sizeof(float *) );
   assert(M->row);
   M->data = buf = (float *)shmalloc( nr * nc * sizeof(float) );
   assert(M->data);
  }
 M->nrows = nr;
 M->ncols = nc;
 farm
   forall( i, 0, nr, # )
     M->row[i] = buf + i * nc;
 return M;
}


void printMatrix( Matrix M )
{
 int m = M->nrows;
 int n = M->ncols;
 int i,j;
 printf("Matrix %dX%d:\n", m, n );
 if (m > 8)  return;
 for (i=0; i<m; i++) {
    for (j=0; j<n; j++)
       printf(" %2.2f", Get( M,i,j));
    printf("\n");
 }
 printf("\n");
}


void freeMatrix( Matrix M )    // only for matrices on global shared heap
{
 assert( M );
 shfree(M->data);
 shfree(M->row);
 shfree(M);
}


/** create 4 quarter-submatrix views of M (no copy of data buffer)
 */
sync Matrix **splitMatrix( sh Matrix M, sh int x, sh int y ) 
{
 sh int p = #;
 sh float *buf;
 sh int m, n;
 sh Matrix **Q;
 int i;
 
 seq {
   assert( M );
   assert( x > 0 );
   assert( y > 0 );
   m = M->nrows;
   n = M->ncols;
   assert( x < m );
   assert( y < n );
   pprintf("splitMatrix( %dx%d -> %dx%d etc. with %d pr\n", m, n, x, y, p );
   Q = (Matrix **)shmalloc( 2*sizeof( Matrix *));
   Q[0] = (Matrix *)shmalloc( 2*sizeof( Matrix ));
   Q[1] = (Matrix *)shmalloc( 2*sizeof( Matrix ));
 }
 farm
   forall( i, 0, 4, p )
     Q[i/2][i%2] = (Matrix)shmalloc( sizeof(matrix) );
 seq {
   Q[0][0]->nrows = x;
   Q[0][0]->ncols = y;
   Q[0][0]->row = (float **)shmalloc( x * sizeof(float *) );
   Q[0][1]->nrows = x;
   Q[0][1]->ncols = n-y;
   Q[0][1]->row = (float **)shmalloc( x * sizeof(float *) );
   Q[1][0]->nrows = m-x;
   Q[1][0]->ncols = y;   
   Q[1][0]->row = (float **)shmalloc( (m-x) * sizeof(float *) );
   Q[1][1]->nrows = m-x;
   Q[1][1]->ncols = n-y;
   Q[1][1]->row = (float **)shmalloc( (m-x) * sizeof(float *) );
 }
 farm {
   forall( i, 0, x, p ) {
      Q[0][0]->row[i] = M->row[i];
      Q[0][1]->row[i] = M->row[i] + y;
   }
   forall( i, 0, m-x, p ) {
      Q[1][0]->row[i] = M->row[i+x];
      Q[1][1]->row[i] = M->row[i+x] + y;
   }
 }
 return Q;
}




sync Matrix *oldsplitMatrix( sh Matrix M, int x, int y ) 
{
 // create 4 quarter-submatrix views of M (no copy of data buffer)
 sh Matrix *sM;
 sh int p = groupsize();
 sh float *buf;
 sh int m, n;
 int i;
 
 seq {
   assert( M );
   assert( x > 0 );
   assert( y > 0 );
   m = M->nrows;
   n = M->ncols;
   assert( x < m );
   assert( y < n );
   //pprintf("splitMatrix( %dx%d -> %dx%d etc.\n", m, n, x, y );
   sM = (Matrix *)shmalloc( 4 * sizeof(Matrix) );
 }
 farm
   forall( i, 0, 4, p )
     sM[i] = (Matrix)shmalloc( sizeof(matrix) );
 seq {
   sM[0]->nrows = x;
   sM[0]->ncols = y;
   sM[0]->row = (float **)shmalloc( x * sizeof(float *) );
   // sM[0]->data = M->data;
   sM[1]->nrows = x;
   sM[1]->ncols = n-y;
   sM[1]->row = (float **)shmalloc( x * sizeof(float *) );
   // sM[1]->data = M->data;
   sM[2]->nrows = m-x;
   sM[2]->ncols = y;
   sM[2]->row = (float **)shmalloc( (m-x) * sizeof(float *) );
   // sM[2]->data = M->data;
   sM[3]->nrows = m-x;
   sM[3]->ncols = n-y;
   sM[3]->row = (float **)shmalloc( (m-x) * sizeof(float *) );
   // sM[3]->data = M->data;
 }
 farm {
   forall( i, 0, x, p ) {
      sM[0]->row[i] = M->row[i];
      sM[1]->row[i] = M->row[i] + y;
   }
   forall( i, 0, m-x, p ) {
      sM[2]->row[i] = M->row[i+x];
      sM[3]->row[i] = M->row[i+x] + y;
   }
 }
 return sM;
}



sync void Add( sh Matrix A, sh Matrix B, sh Matrix C )
{
 sh int p = groupsize();
 sh int n, m;
 int ij, i, j;

 farm {
   assert(A);
   assert(B);
   assert(C);
   n = A->nrows;
   m = A->ncols;
   pprintf("Add %p:%dx%d, %p:%dx%d => %p:%dx%d with %d pr\n", A, n,m, 
            B, B->nrows, B->ncols, C, C->nrows, C->ncols, p );
   assert( n == B->nrows );
   assert( n == C->nrows );
   assert( m == B->ncols );
   assert( m == C->ncols );
   forall2( i, j, ij, 0, n, 0, m, p ) {
     //i = ij / n;
     //j = ij % n;
     Set( C, i,j, Get(A,i,j) + Get(B,i,j) );
   }
 }
}


sync void Sub( sh Matrix A, sh Matrix B, sh Matrix C )
{
 sh int p = groupsize();
 sh int n = A->nrows;
 sh int m = A->ncols;
 int ij, i, j;

 farm pprintf("Sub %p:%dx%d, %p:%dx%d => %p:%dx%d with %d pr\n", A, n,m, 
              B, B->nrows, B->ncols, C, C->nrows, C->ncols, p );
 farm {
   assert( n == B->nrows );
   assert( m == B->ncols );
   forall2( i, j, ij, 0, n, 0, m, p ) {
     //i = ij / n;
     //j = ij % n;
     Set( C, i,j, Get(A,i,j) - Get(B,i,j) );
   }
 }
}


sync void Mul( sh Matrix A, sh Matrix B, sh Matrix C )
{  // standard method
 sh int p = groupsize();
 sh int n, m, r;
 int ij, i, j, k;
 float s;
 seq {
   assert(A);
   assert(B);
   assert(C);
   n = A->nrows;
   m = B->ncols;
   r = A->ncols;
   assert( r == B->nrows );
   assert( n == C->nrows );
   assert( m == C->ncols );
 }
 farm pprintf("Mul %p:%dx%d, %p:%dx%d, => %p:%dx%d with %d pr\n", A, n,m, 
              B, B->nrows, B->ncols, C, C->nrows, C->ncols, p );
 farm {
   forall2( i, j, ij, 0, n, 0, m, p ) {
   //forall( ij, 0, n*m, p ) {}
     //i = ij / m;
     //j = ij % m;
     s = 0.0;
     for (k=0; k<r; k++)
        s += Get(A,i,k) * Get(B,k,j);
     Set( C, i,j, s );
   }
 }
}

