/*************************************************/
/*                                               */
/*      mpi.c                                    */
/*      message passing interface in fork95      */
/*      v0.22                                    */
/*      29.10.1999 Christoph W. Kessler,         */
/*      Oliver Fritzen                           */
/*                                               */
/*************************************************/

#include <fork.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include "mpi.h"

typedef struct msg_struct {
  struct msg_struct *next;
  char *buffer;
  int count;
  MPI_Datatype datatype;
  int tag;
  MPI_Comm comm;
  int messageNr;
} *msg;

typedef struct msgList_struct {
  simple_lock *lock; 
  msg first;
  msg last;
} *msgList;


sh msgList **mailbox;             /* receive buffers */

sh int globalMessageCounter;

sh int whoisfirst = 0;

sh int MEMBER_ARR_SIZE;

sh simple_lock otlk;

pr MPI_Comm MPI_COMM_WORLD;
pr MPI_Group MPI_GROUP_EMPTY;

pr int (*red_func_table[MPI_NR_OF_RED_FUNCS])
  (void *out, void *in, int *count, MPI_Datatype *type);


/*********************************************************/
/*             Begin MPI_Ops Definition                  */
/*********************************************************/

async int mpi_sum(void *out,  void *in, 
		  int  *len, MPI_Datatype *type) {
  /* add over n vectors, returns one sum vector */
  int i;
  
  if((*type >= MPI_CHAR)&&(*type <= MPI_UNSIGNED_LONG)) {
    for(i=0; i < *len; i++) 
      *(((int *) out)+i) += *(((int *) in)+i);
  }
  else if((*type >= MPI_FLOAT)&&(*type <= MPI_LONG_DOUBLE)) {
    for(i=0; i < *len; i++) 
      *(((float *) out)+i) += *(((float *) in)+i);
  }
  /* LONG_LONG_INT not supported */
  return 0;
}

async int mpi_max(void *out,  void *in, 
		  int  *len, MPI_Datatype *type) {
  /* max over n vectors, returns one max vector */
  int i;
  
  if((*type >= MPI_CHAR)&&(*type <= MPI_UNSIGNED_LONG)) {
    for(i=0; i < *len; i++) 
      if(*(((int *) out)+i) < *(((int *) in)+i))
	(*(((int *) out)+i) = *(((int *) in)+i));
  }
  else if((*type >= MPI_FLOAT)&&(*type <= MPI_LONG_DOUBLE)) {
    for(i=0; i < *len; i++) 
      if(*(((float *) out)+i) < *(((float *) in)+i))
	(*(((float *) out)+i) = *(((float *) in)+i));
  }
  /* LONG_LONG_INT not supported */
  return 0;
}
  

async int mpi_min(void *out,  void *in, 
		  int  *len, MPI_Datatype *type) {
  /* min over n vectors, returns one min vector */
  int i;
  if((*type >= MPI_CHAR)&&(*type <= MPI_UNSIGNED_LONG)) {
    for(i=0; i < *len; i++) 
      if(*(((int *) out)+i) > *(((int *) in)+i))
	(*(((int *) out)+i) = *(((int *) in)+i));
  }
  else if((*type >= MPI_FLOAT)&&(*type <= MPI_LONG_DOUBLE)) {
    for(i=0; i < *len; i++) 
      if(*(((float *) out)+i) > *(((float *) in)+i))
	(*(((float *) out)+i) = *(((float *) in)+i));
  }
  /* LONG_LONG_INT not supported */
  return 0;
}

async int mpi_prod(void *out,  void *in, 
		   int  *len, MPI_Datatype *type) {
  /* product over n vectors, returns one product vector */
  int i;
  
  if((*type >= MPI_CHAR)&&(*type <= MPI_UNSIGNED_LONG)) {
    for(i=0; i < *len; i++) 
      *(((int *) out)+i) *= *(((int *) in)+i);
  }
  else if((*type >= MPI_FLOAT)&&(*type <= MPI_LONG_DOUBLE)) {
    for(i=0; i < *len; i++) 
      *(((float *) out)+i) = (*(((float *) out)+i)) * (*(((float *) in)+i));
  }
  /* LONG_LONG_INT not supported */
  return 0;
}


async int mpi_land(void *out,  void *in, 
		   int  *len, MPI_Datatype *type) {
  /* logical and over n vectors, returns one LAND vector */
  int i;
  
  if((*type >= MPI_CHAR)&&(*type <= MPI_UNSIGNED_LONG)) {
    for(i=0; i < *len; i++) 
      *(((int *) out)+i) = *(((int *) out)+i) && *(((int *) in)+i);
  }
  else if((*type >= MPI_FLOAT)&&(*type <= MPI_LONG_DOUBLE)) {
    for(i=0; i < *len; i++) 
      *(((float *) out)+i) = *(((float *) out)+i) && *(((float *) in)+i);
  }
  /* LONG_LONG_INT not supported */
  return 0;
}


async int mpi_band(void *out,  void *in, 
		   int  *len, MPI_Datatype *type) {
  /* bitwise and over n vectors, returns one BAND vector */
  int i;
  
  if((*type >= MPI_CHAR)&&(*type <= MPI_UNSIGNED_LONG)) {
    for(i=0; i < *len; i++) 
      *(((int *) out)+i) = *(((int *) out)+i) & *(((int *) in)+i);
  }
  else return 11; /* op not provided for this type */
  return 0;
}

async int mpi_lor(void *out,  void *in, 
		   int  *len, MPI_Datatype *type) {
  /* logical or over n vectors, returns one LOR vector */
  int i;
  
  if((*type >= MPI_CHAR)&&(*type <= MPI_UNSIGNED_LONG)) {
    for(i=0; i < *len; i++) 
      *(((int *) out)+i) = *(((int *) out)+i) || *(((int *) in)+i);
  }
  else if((*type >= MPI_FLOAT)&&(*type <= MPI_LONG_DOUBLE)) {
    for(i=0; i < *len; i++) 
      *(((float *) out)+i) = *(((float *) out)+i) || *(((float *) in)+i);
  }
  else return 11; /* op not provided for this type */
  return 0;
}

async int mpi_bor(void *out,  void *in, 
		  int  *len, MPI_Datatype *type) {
  /* bitwise or over n vectors, returns one BOR vector */
  int i;
  
  if((*type >= MPI_CHAR)&&(*type <= MPI_UNSIGNED_LONG)) {
    for(i=0; i < *len; i++) 
      *(((int *) out)+i) = *(((int *) out)+i) | *(((int *) in)+i);
  }
  else return 11; /* op not provided for this type */
  return 0;
}

async int mpi_lxor(void *out,  void *in, 
		   int  *len, MPI_Datatype *type) {
  /* logical xor over n vectors, returns one LXOR vector */
  int i;
  
  if((*type >= MPI_CHAR)&&(*type <= MPI_UNSIGNED_LONG))
    for(i=0; i < *len; i++) 
      *(((int *) out)+i) = 
	((*(((int *) out)+i) == 0) && *(((int *) in)+i))
	||
	(*(((int *) out)+i) && (*(((int *) in)+i) == 0));
	
	else return 11; /* Op not provided for this type */
  return 0;
}

async int mpi_bxor(void *out,  void *in, 
		   int  *len, MPI_Datatype *type) {
  /* bitwise or over n vectors, returns one BXOR vector */
  int i;
  
  if(*type == MPI_INT)
    for(i=0; i < *len; i++) *(((int *) out)+i) ^= *(((int *) in)+i);
  else return 11; /* Op not provided for this type */
  return 0;
}

async int mpi_maxloc(void *out,  void *in, 
		     int  *len, MPI_Datatype *type) {
  int i;
  mpi_2int *ia, *ib;
  mpi_float_int *fa, *fb;
  if((*type >= MPI_LONG_INT)&&(*type <= MPI_2INT)) {
    ia = (mpi_2int *) out;
    ib = (mpi_2int *) in;
    for(i=0; i < *len; i++) if(ia[i].var < ib[i].var) {
      ia[i].var = ib[i].var;
      ia[i].loc = ib[i].loc;
    }
    else 
      if(ia[i].var == ib[i].var) 
	if(ia[i].loc > ib[i].loc) ia[i].loc = ib[i].loc;    
  }
  else if(((*type >= MPI_FLOAT_INT)&&(*type <= MPI_DOUBLE_INT))||
	  (*type == MPI_LONG_DOUBLE)) {
    fa = (mpi_float_int *) out;
    fb = (mpi_float_int *) in;
    for(i=0; i < *len; i++) if(fa[i].var < fb[i].var) {
      fa[i].var = fb[i].var;
      fa[i].loc = fb[i].loc;
    }
    else 
      if(fa[i].var == fb[i].var) 
	if(fa[i].loc > fb[i].loc) fa[i].loc = fb[i].loc;
  }
  else return 11; 
  return 0;
}

async int mpi_minloc(void *out,  void *in, 
		     int  *len, MPI_Datatype *type) {
  int i;
  mpi_2int *ia, *ib;
  mpi_float_int *fa, *fb;
  if(*type == MPI_2INT) {
    ia = (mpi_2int *) out;
    ib = (mpi_2int *) in;
    for(i=0; i < *len; i++) if(ia[i].var > ib[i].var) {
      ia[i].var = ib[i].var;
      ia[i].loc = ib[i].loc;
    }
    else 
      if(ia[i].var == ib[i].var) 
	if(ia[i].loc > ib[i].loc) ia[i].loc = ib[i].loc;    
  }
  else if(*type == MPI_FLOAT_INT) {
    fa = (mpi_float_int *) out;
    fb = (mpi_float_int *) in;
    for(i=0; i < *len; i++) if(fa[i].var > fb[i].var) {
      fa[i].var = fb[i].var;
      fa[i].loc = fb[i].loc;
    }
    else 
      if(fa[i].var == fb[i].var) 
	if(fa[i].loc > fb[i].loc) fa[i].loc = fb[i].loc;
  }
  else return 11; 
  return 0;
} 
/*********************************************************/
/*              End MPI_Ops Definition                   */
/*********************************************************/


async int MPI_Barrier(MPI_Comm comm) {
  int size, myrank;
  MPI_Comm_size(comm, &size);
  MPI_Comm_rank(comm, &myrank);
  if(myrank != -1) {
    if(mpadd(comm.comm_barrier, 1) != (size-1)) {  
      while((*(comm.comm_barrier)) != size);
    }
    else {
      while((*(comm.comm_barrier)) != size);
      *(comm.comm_barrier) = 0;
    }
  }
  return 0;
}

async void set_all_bits(int *array, int n) {
  /* 
   * returns int-array with the n last 
   * bits set to 1, all others 0 
   */
  int i;
  for(i=0;  i < (n/(SYSTEM_WORD_LENGTH * sizeof(int)));i++) array[i] = -1;
  array = array+i;
  for(i=0; i < n % (sizeof(int) * SYSTEM_WORD_LENGTH);i++) {
    array[0] = (array[0]) << 1;
    (array[0])++;
  }
}

async void Init_group(MPI_Group *group) {
  group->size = 0;
  group->members = NULL;
  group->ranks = NULL;
}

async int MPI_Send(const void *message, int count, MPI_Datatype datatype,
		   int dest, int tag, MPI_Comm comm) {
  
  msg m;
  
  MPI_Status status;
  MPI_Group group;
  int header_size = sizeof(MPI_Status);
  int myrank;
  
  Init_group(&group);
  MPI_Comm_group(comm, &group);
  MPI_Group_rank(group, &myrank);
  if(myrank == -1) {
    MPI_Group_free(&group);
    return MPI_ERR_COMM;  /* invalid communicator */
  }
  status.count = count;
  status.MPI_SOURCE = myrank;
  status.MPI_TAG = tag;
  status.private_count = 0; /* ??????????? */
  
  m = (msg) shmalloc(sizeof(struct msg_struct)); 
  if(!m) return MPI_ERR_SYSRESOURCE; 
  m->messageNr = mpadd(&globalMessageCounter,1);
  m->count = count;
  m->datatype = datatype;
  m->tag = tag;
  m->comm = comm;
  m->next = NULL;
  m->buffer = (char *) 
    shmalloc(header_size + (count * MPI_Sizeof(datatype)));
  if(!(m->buffer)) return MPI_ERR_SYSRESOURCE;
  memcpy((char *) m->buffer, (char *) (&status), header_size);
  memcpy((char *) (m->buffer+header_size), (char *) message, 
	 count * MPI_Sizeof(datatype));
  simple_lockup(mailbox[group.ranks[dest]][__PROC_NR__]->lock);
  if(mailbox[group.ranks[dest]][__PROC_NR__]->first == NULL) {
    mailbox[group.ranks[dest]][__PROC_NR__]->first = m;
    mailbox[group.ranks[dest]][__PROC_NR__]->last = m;
  }
  else {
    mailbox[group.ranks[dest]][__PROC_NR__]->last->next = m;
    mailbox[group.ranks[dest]][__PROC_NR__]->last = m; 
  }
  simple_unlock(mailbox[group.ranks[dest]][__PROC_NR__]->lock);
  return 0;
}


async int MPI_Bcast(const void *message, int count, MPI_Datatype datatype,
		    int root, MPI_Comm comm) {
  
  int i, size, myrank, dest;
  MPI_Group group;
  MPI_Status status;
  
  Init_group(&group);
  MPI_Comm_size(comm, &size);
  MPI_Comm_group(comm, &group);
  MPI_Comm_rank(comm, &myrank);
  
  if(size < 2) return 0;
  if((root < 0) || (root > size)) return MPI_ERR_COMM;
  if(root) {
    if(myrank == root) MPI_Send(message, count, datatype, 0, -2, comm);
    /* send message from root to proc 0 */
    else if(myrank == 0) 
      MPI_Recv(message, count, datatype, root, -2, comm, &status);
    /* proc 0 waits for message from root */
  }
  if(myrank)
    MPI_Recv(message, count, datatype, (myrank-1)/2, -2, comm, &status);
  if((dest = ((myrank << 1)+1)) < size) 
    MPI_Send(message, count, datatype, dest, -2, comm);
  if((++dest) < size) 
    MPI_Send(message, count, datatype, dest, -2, comm);
  
  MPI_Barrier(comm);
  if(root) {
    if(myrank == 0) 
      MPI_Send(message, count, datatype, root, -2, comm);
    /* Send message from proc zero to root */
    else if(myrank == root)
      MPI_Recv(message, count, datatype, 0, -2, comm, &status);
    /* root receives message from proc zero */  
  }
  MPI_Barrier(comm);
  return 0;
}


async int MPI_Recv(const void *message, int count, MPI_Datatype datatype,
		   int source, int tag, MPI_Comm comm, MPI_Status *status) {
  
  msg m = NULL,
    m_old = NULL;
  msgList mBox;
  
  int header_size = sizeof(MPI_Status);
  int myrank, group_size;
  int **addr;
  MPI_Group group;
  Init_group(&group);
  MPI_Comm_group(comm, &group);
  MPI_Group_size(group, &group_size);
  MPI_Group_rank(group, &myrank);
  
  if(source == MPI_ANY_SOURCE) {
    /* go through all own mailbox slots 
       until the oldest message is found */
    
    int i,index_min = 0;
    msg   m_min = NULL,
      m_min_old = NULL;
    if(myrank == -1) return MPI_ERR_COMM;
    while(! m_min) {
      for(i=0; i<group_size; i++) {
	mBox = mailbox[__PROC_NR__][group.ranks[i]];
	m = mBox->first;
	m_old = NULL;
	if(tag != MPI_ANY_TAG) {
	  /* if tags DO matter */
	  while(m && (m->tag != tag)) {
	    m_old = m;
	    m = m->next;
	  }
	}
	if(m && ((! m_min) || (m->messageNr < m_min->messageNr))) {
	  m_min = m;
	  m_min_old = m_old;
	  index_min = i;
	}
      }
    }
    m = m_min;
    m_old = m_min_old;
    source = index_min;
  }
  else {
    /* go through one specified mailbox slot (source) 
       and retrieve the oldest message from there     */
   
    while(! m) {
      int i;
      mBox = mailbox[__PROC_NR__][group.ranks[source]];
      m = mBox->first;
      if(tag != MPI_ANY_TAG) while(m && (m->tag != tag)) {
	m_old = m;
	m = m->next;
      }
    }
  }
  simple_lockup(mailbox[__PROC_NR__][group.ranks[source]]->lock);
  if(m_old) 
    m_old->next = m->next;
  else mailbox[__PROC_NR__][group.ranks[source]]->first = m->next;
  if(! m->next) 
    mailbox[__PROC_NR__][group.ranks[source]]->last = m_old;
  simple_unlock(mailbox[__PROC_NR__][group.ranks[source]]->lock);
  
  /* divide buffer in header... */
  memcpy(status, m->buffer, header_size);
  /* ...and data part           */
  memcpy(message, (m->buffer + header_size) ,count * MPI_Sizeof(datatype));
  
  /* deallocate list element(m) and its buffer(m->buffer) */
  shfree(m->buffer);
  shfree(m);
  return 0;
}

/******************** group operations ***********************/

async int MPI_Group_union(MPI_Group group1, MPI_Group group2,
			  MPI_Group *new_group) {

  /* Ann.:
   * all group operations can cope with calls of 
   * the following form:
   * MPI_Group_someoperation(group_A, group_B, group_A);
   * will say: groups can be used as source AND as destination
   * parameters at the same time
   */
  
  int i,j,mask,mem_index;
  int *temp_ranks;
  int *temp_members;
  int temp_size = group1.size;
  
  temp_members = (int *) shmalloc(MEMBER_ARR_SIZE * sizeof(int));
  temp_ranks   = (int *) shmalloc(sizeof(int) * (group1.size+group2.size));
  memcpy(temp_ranks,group1.ranks,group1.size);
  
  for(i=0; i<MEMBER_ARR_SIZE; i++)
    /* unify 'members'-array */
    temp_members[i] = group1.members[i] | group2.members[i];
  
  for(i=0; i<group2.size; i++) {
    /* unify 'ranks'-array */
    mem_index = (group2.ranks[i])/(SYSTEM_WORD_LENGTH * sizeof(int));
    mask = 1 << (group2.ranks[i] %(SYSTEM_WORD_LENGTH * sizeof(int)));
    if(group2.members[mem_index] & (~group1.members[mem_index]) & mask)
      temp_ranks[temp_size++] = group2.ranks[i];
  }  
  
  /* delete old fields */
  MPI_Group_free(new_group);
  
  /* assign new ones */
  new_group->members = temp_members;
  new_group->ranks = 
    (int *) shmalloc(temp_size * sizeof(int));
  memcpy(new_group->ranks,   temp_ranks,   temp_size);
  new_group->size = temp_size;
  
  shfree(temp_ranks);   /* free(temp_ranks)   */
  return 0;
}

async int MPI_Group_intersection(MPI_Group group1, MPI_Group group2,
				 MPI_Group *new_group) {
  int i,j,mask,mem_index,
    temp_size=0;
  int *temp_members;
  int *temp_ranks;
  
  temp_members = (int *) /**/shmalloc(MEMBER_ARR_SIZE * sizeof(int));
  temp_ranks   = (int *) /**/shmalloc(group1.size * sizeof(int));
  
  for(i=0; i<MEMBER_ARR_SIZE; i++)
    /* intersect 'members'-array */
    temp_members[i] = group1.members[i] & group2.members[i];
  
  for(i=0; i<group1.size; i++) {
    /* intersect 'ranks'-array */
    mem_index = (group1.ranks[i])/(SYSTEM_WORD_LENGTH * sizeof(int));
    mask = 1 << (group1.ranks[i] %(SYSTEM_WORD_LENGTH * sizeof(int)));
    if(temp_members[mem_index] & mask)
      temp_ranks[temp_size++] = group1.ranks[i];
  }  
  /* now, copy temp values/arrays to new_group */
  MPI_Group_free(new_group);
  new_group->members = temp_members;
  new_group->ranks = 
    (int *) shmalloc(temp_size * sizeof(int));
  memcpy(new_group->ranks, temp_ranks, temp_size);
  new_group->size = temp_size;
  shfree(temp_ranks);   /* free(temp_ranks)   */
  return 0;
}

async int MPI_Group_difference(MPI_Group group1, MPI_Group group2,
			       MPI_Group *new_group) {
  int i,j,mask,mem_index,
    temp_size=0;
  int *temp_members;
  int *temp_ranks;
  
  temp_members = (int *) /**/shmalloc(MEMBER_ARR_SIZE * sizeof(int));
  temp_ranks   = (int *) /**/shmalloc(group1.size * sizeof(int));
  
  for(i=0; i<MEMBER_ARR_SIZE; i++)
    /* intersect 'members'-array */
    temp_members[i] = group1.members[i] & (~group2.members[i]);
  
  for(i=0; i<group1.size; i++) {
    /* intersect 'ranks'-array */
    mem_index = (group1.ranks[i])/(SYSTEM_WORD_LENGTH * sizeof(int));
    mask = 1 << (group1.ranks[i] %(SYSTEM_WORD_LENGTH * sizeof(int)));
    if(temp_members[mem_index] & mask)
      temp_ranks[temp_size++] = group1.ranks[i];
  }  
  /* now, copy temp values/arrays to new_group */
  MPI_Group_free(new_group);
  new_group->members = temp_members;
  new_group->ranks = 
    (int *) shmalloc(temp_size * sizeof(int));
  memcpy(new_group->ranks, temp_ranks, temp_size);
  new_group->size = temp_size;
  shfree(temp_ranks);   /* free(temp_ranks)   */
  return 0;  
}

async int MPI_Group_incl(MPI_Group group, int n, int *ranks, 
			 MPI_Group *new_group) {

  int i,mem_index,mask,
    *temp_members,
    *temp_ranks;
  
  if((! group.ranks)||(! group.members)) return MPI_ERR_GROUP;
  if(n > group.size) return MPI_ERR_ARG;
  
  temp_members = (int *) /**/ shmalloc(MEMBER_ARR_SIZE * sizeof(int));
  temp_ranks = (int *) /**/ shmalloc(n * sizeof(int));
  for(i=0; i<MEMBER_ARR_SIZE; i++) temp_members[i] = 0;
  for(i=0; i<n ; i++) {
    /* build new 'ranks'-array */
    mem_index = (group.ranks[ranks[i]])/(SYSTEM_WORD_LENGTH * sizeof(int));
    mask = 1 << (group.ranks[ranks[i]] %(SYSTEM_WORD_LENGTH * sizeof(int)));
    temp_members[mem_index] |= mask;
    temp_ranks[i] = group.ranks[ranks[i]];
  }  
  /* now, copy temp values/arrays to new_group */
  MPI_Group_free(new_group);
  new_group->members = temp_members;
  new_group->ranks = temp_ranks;
  new_group->size = n;
  return 0;
}


async int MPI_Group_excl(MPI_Group group, int n, int *ranks, 
			 MPI_Group *new_group) {
  int i, mem_index, mask;
  int *temp_members;
  int *temp_ranks;
  MPI_Group diff; 
  
  if((!group.ranks)||(!group.members)) return MPI_ERR_GROUP;
  if(n > group.size) return MPI_ERR_ARG;
  
  diff.size = n;
  diff.members = (int *) /**/shmalloc(MEMBER_ARR_SIZE * sizeof(int));
  diff.ranks   = (int *) /**/shmalloc(n * sizeof(int));
  
  for(i=0; i<MEMBER_ARR_SIZE; i++) diff.members[i] = 0;
  for(i=0; i<n; i++) {
    /* build new 'ranks'-array */
    mem_index = (group.ranks[ranks[i]])/(SYSTEM_WORD_LENGTH * sizeof(int));
    mask = 1 << (group.ranks[ranks[i]] %(SYSTEM_WORD_LENGTH * sizeof(int)));
    diff.members[mem_index] |= mask;
    diff.ranks[i] = group.ranks[ranks[i]];
  }
  MPI_Group_difference(group, diff, new_group);
  MPI_Group_free(&diff);
  return 0;
}

async int MPI_Group_size(MPI_Group group, int *size) {
  *size = group.size;
  return 0;
}


async int MPI_Group_rank(MPI_Group group, int *rank) {
  int i=0;
  *rank = -1;
  while(i < group.size) {
    if(group.ranks[i] == __PROC_NR__) {
      *rank = i;
      break;
    }
    i++;
  }
  return 0;
}

async int MPI_Group_compare(MPI_Group g1, MPI_Group g2, int *res) {
  /* since topologies are not (yet) supported, 
     compare won't EVER return 'congruent'     */
  
  int i,min;
  *res = MPI_IDENT;
  if((min = g1.size) != g2.size) {
    *res = MPI_UNEQUAL;
    return 0;
  }
  for(i=0; i<MEMBER_ARR_SIZE; i++) {
    if(g1.members[i] != g2.members[i]) {
      *res = MPI_UNEQUAL;
      return 0;
    }
  }
  for(i=0; i<g1.size; i++) {
    if(g1.ranks[i] != g2.ranks[i]) {
      *res = MPI_SIMILAR;
      return 0;
    }
  }
  *res = MPI_IDENT;
  return 0;
}


async int MPI_Group_free(MPI_Group *group) {
  if(group) {
    if(group->members) {
      shfree(group->members);
      group->members = NULL;
    }
    if(group->ranks) {
      shfree(group->ranks);
      group->ranks = NULL;
    }
    group->size = 0;
    return 0;
  }
  else  {
    return MPI_ERR_GROUP;
  }
}

async int MPI_Comm_group(MPI_Comm comm, MPI_Group *group) {
  MPI_Group_free(group);
  group->size = comm.group.size;
  group->members = (int *) shmalloc(MEMBER_ARR_SIZE * sizeof(int));
  memcpy(group->members, comm.group.members, MEMBER_ARR_SIZE * sizeof(int));
  group->ranks = (int *) shmalloc(group->size * sizeof(int));
  memcpy(group->ranks, comm.group.ranks, group->size * sizeof(int));
  return 0;
}

async int MPI_Comm_create(MPI_Comm comm, 
			  MPI_Group group, 
			  MPI_Comm *new_comm) {
  /* 
   * error values:
   * 11:     group isn't subset of comm.group 
   */
  
  int i,myrank,size;
  
  MPI_Group_size(group, &size);
  MPI_Group_rank(group, &myrank);
  
  if(! &comm) return MPI_ERR_COMM;
  for(i=0; i<MEMBER_ARR_SIZE; i++)
    if((comm.group.members[i] | group.ranks[i]) 
       > comm.group.members[i]) 
      return MPI_ERR_ARG;
  new_comm->group.size = group.size;
  new_comm->group.members= (int *) shmalloc(MEMBER_ARR_SIZE * sizeof(int));
  memcpy(new_comm->group.members, group.members,MEMBER_ARR_SIZE * sizeof(int));
  new_comm->group.ranks = (int *) shmalloc(group.size * sizeof(int));
  memcpy(new_comm->group.ranks, group.ranks,comm.group.size * sizeof(int));
  if(! myrank) {
    new_comm->comm_barrier = (int *) shmalloc(sizeof(int));
    *(new_comm->comm_barrier) = 0;
    for(i=1; i<size; i++) 
      MPI_Send(&(new_comm->comm_barrier), 1, MPI_INT, i, -4, *new_comm);
  }
  else if(myrank > 0) {
    MPI_Status status;
    MPI_Recv(&(new_comm->comm_barrier), 1, MPI_INT, 0, -4, *new_comm, &status);
  }

  /*********************************/
  if(! myrank) {
    *(new_comm->comm_barrier) = 0;
    for(i=1; i<size; i++) 
      MPI_Send(NULL, 1, MPI_INT, i, -6, *new_comm);
  }
  else if(myrank > 0) {
    void *foobar;
    MPI_Status status;
    MPI_Recv(foobar, 1, MPI_INT, 0, -6, *new_comm, &status);
  }
  /* workaround: set comm_barrier to zero,
   * after that synchronize group with Recv
   */
  return 0;
}

async int MPI_Comm_free(MPI_Comm *comm) {
  MPI_Group_free(&(comm->group));
  shfree(comm->comm_barrier);
}


async int MPI_Comm_size(MPI_Comm comm, int *size) {
  MPI_Group group;
  Init_group(&group);
  MPI_Comm_group(comm, &group);
  MPI_Group_size(group, size);
  return 0;
}

/* not yet tested! */
async int MPI_Comm_rank(MPI_Comm comm, int *rank) {
  MPI_Group group;
  Init_group(&group);
  MPI_Comm_group(comm, &group);
  MPI_Group_rank(group, rank);
  return 0;
}

async int MPI_Comm_dup(MPI_Comm comm, MPI_Comm *new_comm) {
  return MPI_Comm_create(comm, comm.group, new_comm);
}

async int MPI_Comm_split(MPI_Comm comm, 
			 int color, int key, MPI_Comm *new_comm) {
  /* not yet implemented */
}





async int MPI_Init(int *argc, char ***argv) {
  int i,mask,ret = 0;
  int *semaphore = 0;
  
  simple_lock_init(&otlk);
  MEMBER_ARR_SIZE = 1 + (__STARTED_PROCS__-1) / (SYSTEM_WORD_LENGTH * sizeof(int));
  if(mpadd(&whoisfirst,1) == 0) { 
    /* to make sure that only ONE P. does this */
    mailbox = (msgList **) shmalloc( __STARTED_PROCS__ * sizeof(msgList **)); 
    /* allocate mailbox array */
    if (!mailbox) ret = MPI_ERR_INTERN;
    globalMessageCounter = 0;
  }
  barrier;
  if(ret) return ret;

  /* initialize each proc's mailbox */ 
  mailbox[__PROC_NR__] =  	 
    (msgList *) shmalloc(__STARTED_PROCS__ * sizeof(msgList));
  if(! mailbox[__PROC_NR__]) return MPI_ERR_INTERN;
  for(i=0; i<__STARTED_PROCS__; i++) {
    /* allocate message list heads */
    mailbox[i][__PROC_NR__] = 
      (msgList) shmalloc (sizeof(struct msgList_struct));
    mailbox[i][__PROC_NR__]->lock = (simple_lock *) new_SimpleLock();
    mailbox[i][__PROC_NR__]->first = NULL;
    mailbox[i][__PROC_NR__]->last = NULL;     
  }  
  /* Init GROUP_EMPTY */
  MPI_GROUP_EMPTY.size = 0;
  MPI_GROUP_EMPTY.members = (int *) /**/
    shmalloc(MEMBER_ARR_SIZE * sizeof(int));
  for(i=0; i<MEMBER_ARR_SIZE; i++) MPI_GROUP_EMPTY.members[i] = 0; 
  MPI_GROUP_EMPTY.ranks = NULL;
  
  /* Init COMM_WORLD */
  MPI_COMM_WORLD.group.size = __STARTED_PROCS__;
  MPI_COMM_WORLD.group.members = 
    (int *) /**/shmalloc(MEMBER_ARR_SIZE * sizeof(int));
  set_all_bits(MPI_COMM_WORLD.group.members,__STARTED_PROCS__);
  MPI_COMM_WORLD.group.ranks = 
    (int *) /**/shmalloc(__STARTED_PROCS__ * sizeof(int));
  for(i=0; i < __STARTED_PROCS__; i++) 
    MPI_COMM_WORLD.group.ranks[i] = i;
  if(!__PROC_NR__) {
    semaphore = (int *) shmalloc(sizeof(int));
    *semaphore = 0;
    for(i=1; i<__STARTED_PROCS__; i++) 
      MPI_Send(&semaphore, 1, MPI_INT, i, -5, MPI_COMM_WORLD);
  }
  else {
    MPI_Status status;
    MPI_Recv(&semaphore, 1, MPI_INT, 0, -5, MPI_COMM_WORLD, &status);
  }
  MPI_COMM_WORLD.comm_barrier = semaphore;
  /* Init red_func_table */
  red_func_table[MPI_MAX] = &mpi_max;
  red_func_table[MPI_MIN] = &mpi_min;
  red_func_table[MPI_SUM] = &mpi_sum;
  red_func_table[MPI_PROD] = &mpi_prod;
  red_func_table[MPI_LAND] = &mpi_land;
  red_func_table[MPI_BAND] = &mpi_band;
  red_func_table[MPI_BOR] = &mpi_bor;
  red_func_table[MPI_BXOR] = &mpi_bxor;
  red_func_table[MPI_LOR] = &mpi_lor;
  red_func_table[MPI_LXOR] = &mpi_lxor;
  red_func_table[MPI_MAXLOC] = &mpi_maxloc;
  red_func_table[MPI_MINLOC] = &mpi_minloc;
  
  barrier;
  return ret;
}

async int MPI_Finalize(void) {
  int i;
  msg m, old_m;
  
  barrier;
  for(i=0; i < __STARTED_PROCS__; i++) {
    m = mailbox[i][__PROC_NR__]->first; 
    /* empty buffer lists */
    while(m) { 
      old_m = m;
      m = m->next;
      shfree(old_m->buffer);
      shfree(old_m);
    } 
    /* deallocate message list heads */
    shfree(mailbox[i][__PROC_NR__]);
  }
  /* deallocate each proc's mailbox */
  shfree(mailbox[__PROC_NR__]);
  barrier;
  if(__PROC_NR__ == 0) {
    shfree(mailbox);
    shfree(MPI_COMM_WORLD.comm_barrier);
  }
  return 0;
}

async int MPI_Sizeof(MPI_Datatype type) {
  if((type >= MPI_CHAR) && (type <= MPI_LONG_LONG_INT)) return 1;
  if((type >= MPI_FLOAT_INT) && (type <= MPI_LONG_DOUBLE_INT)) 
    return sizeof(mpi_2int);
  /* Obviously only valid for SB-PRAM systems... */
}

async int MPI_Gather(void *sendbuf, int sendcount, MPI_Datatype sendtype,
		     void *recvbuf, int recvcount, MPI_Datatype recvtype, 
		     int root, MPI_Comm comm) {
  /* uses -7 tag for intra communication */
  int i,myrank,size;
  MPI_Status status;
  
  MPI_Comm_rank(comm, &myrank);
  MPI_Comm_size(comm, &size);
  
  if(myrank == -1) return MPI_ERR_COMM;
  if(recvcount > sendcount) return MPI_ERR_ARG;
  if(myrank == root) {
    for(i=0; i < root; i++) 
      MPI_Recv((int *) recvbuf+((i*recvcount*MPI_Sizeof(recvtype))),
	       recvcount,recvtype, i, -7, comm, &status);
    memcpy((int *) recvbuf+((root*recvcount*MPI_Sizeof(recvtype))), 
	   sendbuf, recvcount*MPI_Sizeof(recvtype));
    for(i=root+1; i < size; i++) 
      MPI_Recv((int *) recvbuf+(i*recvcount*MPI_Sizeof(recvtype)),
	       recvcount,recvtype, i, -7, comm, &status);
    
  }
  else MPI_Send(sendbuf, sendcount, sendtype, root, -7, comm);
  MPI_Barrier(comm);
  return 0;
}

async int MPI_Allgather(void *sendbuf, int sendcount, MPI_Datatype sendtype,
			void *recvbuf, int recvcount, MPI_Datatype recvtype, 
			MPI_Comm comm) {
  int size;
  MPI_Comm_size(comm, &size);

  MPI_Gather(sendbuf, sendcount, sendtype,
	     recvbuf, recvcount, recvtype, 
	     0, comm);
  MPI_Bcast(recvbuf, recvcount * size, recvtype, 0, comm);
  return 0;
}

async int MPI_Scatter(void *sendbuf, int sendcount, MPI_Datatype sendtype,
		      void *recvbuf, int recvcount, MPI_Datatype recvtype, 
		      int root, MPI_Comm comm) {
  /* uses -9 tag for intra communication */
  int i,myrank,size;
  MPI_Status status;
  
  MPI_Comm_rank(comm, &myrank);
  MPI_Comm_size(comm, &size);
  
  if(myrank == -1) return MPI_ERR_COMM;
  if(recvcount > sendcount) return MPI_ERR_ARG;
  if(myrank == root) {
    for(i=0; i<root; i++) 
      MPI_Send((int *) sendbuf+(i*sendcount*MPI_Sizeof(sendtype)),
	       sendcount, sendtype, i, -9, comm);
    memcpy(recvbuf, (int *) sendbuf+(root*sendcount*MPI_Sizeof(sendtype)),
	   recvcount*MPI_Sizeof(recvtype));
    for(i=root+1; i<size; i++) 
      MPI_Send((int *) sendbuf+(i*sendcount*MPI_Sizeof(sendtype)),
	       sendcount, sendtype, i, -9, comm);
    
  }
  else MPI_Recv(recvbuf, 5, recvtype, root, -9, MPI_COMM_WORLD, &status);
  MPI_Barrier(comm);
  return 0;
}
  
async int MPI_Reduce(void *sendbuf, void *recvbuf, int count,
 		     MPI_Datatype datatype, MPI_Op op, int root,
 		     MPI_Comm comm) {
  
  /* uses tag -1 for intra-communication */
  
  void *buf;
  int i,size,myrank, softrank, depth, pred, succ;
  
  MPI_Group group;
  MPI_Status status;
  
  Init_group(&group);
  MPI_Comm_group(comm, &group);
  MPI_Group_size(group,&size);
  MPI_Group_rank(group, &myrank);
  
  softrank = myrank;
  depth = 0;
  buf = (void *) shmalloc(count * sizeof(datatype));
  while((softrank >= 0) && (size > 1)) {
    if(softrank % 2) {
      /* send message to predecessor (with even number) and quit */
      pred = myrank - (1<<depth);
      MPI_Send(sendbuf, count, datatype, pred,-1, comm);
      if(myrank == root) softrank = -3; else softrank = -1;
    }
    else {
      /* receive message from successor (with odd number) */
      if(softrank < size-1) {
	/* at least if there IS a successor... */
	succ = myrank + (1<<depth);
	MPI_Recv(buf, count, datatype, succ,
		 -1,comm, &status);
	(red_func_table[op])(sendbuf, buf, &count, &datatype);
      }
      softrank = softrank / 2;
      size = size - (size / 2);
      depth++;
    }
  }
  if(root) {
    if(softrank == -3) {
      MPI_Recv(recvbuf, count, datatype, 0, -1, comm, &status);
    }
    else if(myrank == 0) {
      MPI_Send(sendbuf, count, datatype, root, -1, comm);
    }
  }
  else {
    memcpy(recvbuf, sendbuf, count);
  }
  shfree(buf);
  return 0;
}


/*
 * for debugging purposes ... 
 */
async void print_group(MPI_Group group) {
  int i;
  int size;
  MPI_Group_size(group, &size);
  printf("\nsize   :%4d\n\nmembers:",size);
  if(group.members) for(i=0; i<MEMBER_ARR_SIZE;i++) 
    printf("%4d ", group.members[i]);
  else printf("(no members field)");
  prS("\n\nranks  :");
  if(group.ranks) for(i=0; i<group.size; i++) 
    printf("r[%d] = %d   ", i,group.ranks[i]);
  else printf("(no ranks field)");
  prS("\n");
}







