/*************************************************/
/*                                               */
/*      mpi.c                                    */
/*      message passing interface in fork95      */
/*      v0.15                                    */
/*      30.09.1999 Christoph W. Kessler,         */
/*      Oliver Fritzen                           */
/*                                               */
/*      Notes: - first version of Reduce         */
/*               only MPI_SUM, hardwired         */
/*                                               */
/*                                               */
/*                                               */
/*************************************************/

#include <fork.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include "mpi.h" 
#include <assert.h>

typedef struct msg_struct {
  struct msg_struct *next;
  char *buffer;
  int count;
  MPI_Datatype datatype;
  int tag;
  MPI_Comm comm;
  int messageNr;
} *msg;

typedef struct msgList_struct {
  simple_lock *lock; 
  msg first;
  msg last;
} *msgList;


sh msgList **mailbox;             /* receive buffers */

sh int globalMessageCounter;

sh int whoisfirst = 0;

sh int MEMBER_ARR_SIZE;

pr MPI_Comm MPI_COMM_WORLD;
pr MPI_Group MPI_GROUP_EMPTY;



/* return error codes for: MPI_Send, MPI_Recv, MPI_Comm_size, 
 *                         MPI_Comm_rank, MPI_Init, MPI_Finalize:
 *
 *       General:
 *           0 : no error occured
 *           
 *       Send errors:
 *           11: out of memory/shmalloc error
 *           ...
 *
 *       Recv errors: 
 *           21: out of memory/shmalloc error
 *
 */ 


async void set_all_bits(int *array, int n) {
  /* 
   * returns int-array with the n last 
   * bits set to 1, all others 0 
   */
  int i;
  for(i=0;  i < (n/(SYSTEM_WORD_LENGTH * sizeof(int)));i++) array[i] = -1;
  array = array+i;
  for(i=0; i < n % (sizeof(int) * SYSTEM_WORD_LENGTH);i++) {
    array[0] = (array[0]) << 1;
    (array[0])++;
  }
}

async void Init_group(MPI_Group *group) {
  group->size = 0;
  group->members = NULL;
  group->ranks = NULL;
}

async int MPI_Send(const void *message, int count, MPI_Datatype datatype,
		   int dest, int tag, MPI_Comm comm) {
  
  msg m;
  
  MPI_Status status;
  MPI_Group group;
  int header_size = sizeof(MPI_Status);
  int myrank;
  Init_group(&group);
  MPI_Comm_group(comm, &group);
  MPI_Group_rank(group, &myrank);
  if(myrank == -1) {
    MPI_Group_free(&group);
    return 31; /* I'm not in the communicator */
  }
  status.count = count;
  status.MPI_SOURCE = __PROC_NR__;
  status.MPI_TAG = tag;
  status.private_count = 0; /* ??????????? */
  
  m = (msg) shmalloc(sizeof(struct msg_struct)); 
  if(!m) return 11; 
  m->messageNr = mpadd(&globalMessageCounter,1);
  m->count = count;
  m->datatype = datatype;
  m->tag = tag;
  m->comm = comm;
  m->next = NULL;
  m->buffer = (char *) 
    shmalloc(header_size + (count * MPI_Sizeof(datatype)));
  if(!(m->buffer)) return 11;
  memcpy((char *) m->buffer, (char *) (&status), header_size);
  memcpy((char *) (m->buffer+header_size), (char *) message, count);
  simple_lockup(mailbox[group.ranks[dest]][__PROC_NR__]->lock);
  if(mailbox[group.ranks[dest]][__PROC_NR__]->first == NULL) {
    mailbox[group.ranks[dest]][__PROC_NR__]->first = m;
    mailbox[group.ranks[dest]][__PROC_NR__]->last = m;
  }
  else {
    mailbox[group.ranks[dest]][__PROC_NR__]->last->next = m;
    mailbox[group.ranks[dest]][__PROC_NR__]->last = m; 
  }
  simple_unlock(mailbox[group.ranks[dest]][__PROC_NR__]->lock);
  return 0;
}



/*
 * the following Bcast-implementation is wrong:
 * the syntax is not as given by the MPI forum 
 * and the op is non-blocking 
 * (so stop using it or start fixing it) 
 */

async int MPI_Bcast(const void *message, int count, MPI_Datatype datatype,
		    int tag, MPI_Comm comm) {
  int i, size, dest, ret;
  MPI_Group group;
  MPI_Comm_size(comm, &size);
  MPI_Comm_group(comm, &group);
  for(i=0; i<size; i++) {
    if(i != __PROC_NR__) {
      ret = MPI_Send(message, count, datatype, tag, group.ranks[i], comm);
    }
  }
  return ret;
}


async int MPI_Recv(const void *message, int count, MPI_Datatype datatype,
		   int source, int tag, MPI_Comm comm, MPI_Status *status) {
  
  msg m = NULL,
    m_old = NULL;
  msgList mBox;
  
  int header_size = sizeof(MPI_Status);
  int myrank, group_size;
  MPI_Group group;
  Init_group(&group);
  MPI_Comm_group(comm, &group);
  MPI_Group_size(group, &group_size);
  MPI_Group_rank(group, &myrank);
  
  if(source == MPI_ANY_SOURCE) {
    /* go through all own mailbox slots 
       until the oldest message is found */
    
    int i,index_min = 0;
    msg   m_min = NULL,
      m_min_old = NULL;
    if(myrank == -1) return 11;
    while(! m_min) {
      for(i=0; i<group_size; i++) {
	mBox = mailbox[__PROC_NR__][group.ranks[i]];
	m = mBox->first;
	m_old = NULL;
	if(tag != MPI_ANY_TAG) {
	  /* if tags DO matter */
	  while(m && (m->tag != tag)) {
	    m_old = m;
	    m = m->next;
	  }
	}
	if(m && ((! m_min) || (m->messageNr < m_min->messageNr))) {
	  m_min = m;
	  m_min_old = m_old;
	  index_min = i;
	}
      }
    }
    m = m_min;
    m_old = m_min_old;
    source = index_min;
  }
  else {
    /* go through one specified mailbox slot (source) 
       and retrieve the oldest message from there     */
    while(! m) {
      int i;
      for(i=0; i<group_size; i++) 
	mBox = mailbox[__PROC_NR__][group.ranks[source]];
      m = mBox->first;
      if(tag != MPI_ANY_TAG) while(m && (m->tag != tag)) {
	m_old = m;
	m = m->next;
      }
    }
  }
  simple_lockup(mailbox[__PROC_NR__][group.ranks[source]]->lock);
  if(m_old) 
    m_old->next = m->next;
  else mailbox[__PROC_NR__][group.ranks[source]]->first = m->next;
  if(! m->next) 
    mailbox[__PROC_NR__][group.ranks[source]]->last = m_old;
  simple_unlock(mailbox[__PROC_NR__][group.ranks[source]]->lock);
  
  /* divide buffer in header... */
  memcpy(status, m->buffer, header_size);
  /* ...and data part           */
  memcpy(message, (m->buffer + header_size) ,count);
  assert( source == status->MPI_SOURCE );         //  <--- BUG REPORT
							     
  /* deallocate list element(m) and its buffer(m->buffer) */
  shfree(m->buffer);
  shfree(m);
  return 0;
}

async int MPI_Group_union(MPI_Group group1, MPI_Group group2,
			  MPI_Group *new_group) {

  /* Ann.:
   * all group operations can cope with calls of 
   * the following form:
   * MPI_Group_someoperation(group_A, group_B, group_A);
   * will say: groups can be used as source AND as destination
   * parameters at the same time
   */
  
  int i,j,mask,mem_index;
  int *temp_ranks;
  int *temp_members;
  int temp_size = group1.size;
  
  temp_members = (int *) /**/shmalloc(MEMBER_ARR_SIZE * sizeof(int));
  temp_ranks   = (int *) /**/shmalloc(sizeof(int) * (group1.size+group2.size));
  memcpy(temp_ranks,group1.ranks,group1.size);
  
  for(i=0; i<MEMBER_ARR_SIZE; i++)
    /* unify 'members'-array */
    temp_members[i] = group1.members[i] | group2.members[i];
  
  for(i=0; i<group2.size; i++) {
    /* unify 'ranks'-array */
    mem_index = (group2.ranks[i])/(SYSTEM_WORD_LENGTH * sizeof(int));
    mask = 1 << (group2.ranks[i] %(SYSTEM_WORD_LENGTH * sizeof(int)));
    if(group2.members[mem_index] & (~group1.members[mem_index]) & mask)
      temp_ranks[temp_size++] = group2.ranks[i];
  }  
  
  /* delete old fields */
  MPI_Group_free(new_group);
  
  /* assign new ones */
  new_group->members = temp_members;
  new_group->ranks = 
    (int *) shmalloc(temp_size * sizeof(int));
  memcpy(new_group->ranks,   temp_ranks,   temp_size);
  new_group->size = temp_size;
  
  shfree(temp_ranks);   /* free(temp_ranks)   */
  return 0;
}



async int MPI_Group_intersection(MPI_Group group1, MPI_Group group2,
				 MPI_Group *new_group) {
  int i,j,mask,mem_index,
    temp_size=0;
  int *temp_members;
  int *temp_ranks;
  
  temp_members = (int *) /**/shmalloc(MEMBER_ARR_SIZE * sizeof(int));
  temp_ranks   = (int *) /**/shmalloc(group1.size * sizeof(int));
  
  for(i=0; i<MEMBER_ARR_SIZE; i++)
    /* intersect 'members'-array */
    temp_members[i] = group1.members[i] & group2.members[i];
  
  for(i=0; i<group1.size; i++) {
    /* intersect 'ranks'-array */
    mem_index = (group1.ranks[i])/(SYSTEM_WORD_LENGTH * sizeof(int));
    mask = 1 << (group1.ranks[i] %(SYSTEM_WORD_LENGTH * sizeof(int)));
    if(temp_members[mem_index] & mask)
      temp_ranks[temp_size++] = group1.ranks[i];
  }  
  /* now, copy temp values/arrays to new_group */
  MPI_Group_free(new_group);
  new_group->members = temp_members;
  new_group->ranks = 
    (int *) shmalloc(temp_size * sizeof(int));
  memcpy(new_group->ranks, temp_ranks, temp_size);
  new_group->size = temp_size;
  shfree(temp_ranks);   /* free(temp_ranks)   */
  return 0;
}

async int MPI_Group_difference(MPI_Group group1, MPI_Group group2,
			       MPI_Group *new_group) {
  int i,j,mask,mem_index,
    temp_size=0;
  int *temp_members;
  int *temp_ranks;
  
  temp_members = (int *) /**/shmalloc(MEMBER_ARR_SIZE * sizeof(int));
  temp_ranks   = (int *) /**/shmalloc(group1.size * sizeof(int));
  
  for(i=0; i<MEMBER_ARR_SIZE; i++)
    /* intersect 'members'-array */
    temp_members[i] = group1.members[i] & (~group2.members[i]);
  
  for(i=0; i<group1.size; i++) {
    /* intersect 'ranks'-array */
    mem_index = (group1.ranks[i])/(SYSTEM_WORD_LENGTH * sizeof(int));
    mask = 1 << (group1.ranks[i] %(SYSTEM_WORD_LENGTH * sizeof(int)));
    if(temp_members[mem_index] & mask)
      temp_ranks[temp_size++] = group1.ranks[i];
  }  
  /* now, copy temp values/arrays to new_group */
  MPI_Group_free(new_group);
  new_group->members = temp_members;
  new_group->ranks = 
    (int *) shmalloc(temp_size * sizeof(int));
  memcpy(new_group->ranks, temp_ranks, temp_size);
  new_group->size = temp_size;
  shfree(temp_ranks);   /* free(temp_ranks)   */
  return 0;  
}

async int MPI_Group_incl(MPI_Group group, int n, int *ranks, 
			 MPI_Group *new_group) {

  /* return 11: empty group, can't build subgroup */
  /* return 12: subgroup greater than main group  */
  
  int i,mem_index,mask,
    *temp_members,
    *temp_ranks;
  
  if((! group.ranks)||(! group.members)) return 11;
  if(n > group.size) return 12;
  
  temp_members = (int *) /**/ shmalloc(MEMBER_ARR_SIZE * sizeof(int));
  temp_ranks = (int *) /**/ shmalloc(n * sizeof(int));
  for(i=0; i<MEMBER_ARR_SIZE; i++) temp_members[i] = 0;
  for(i=0; i<n ; i++) {
    /* build new 'ranks'-array */
    mem_index = (group.ranks[ranks[i]])/(SYSTEM_WORD_LENGTH * sizeof(int));
    mask = 1 << (group.ranks[ranks[i]] %(SYSTEM_WORD_LENGTH * sizeof(int)));
    temp_members[mem_index] |= mask;
    temp_ranks[i] = group.ranks[ranks[i]];
  }  
  /* now, copy temp values/arrays to new_group */
  MPI_Group_free(new_group);
  new_group->members = temp_members;
  new_group->ranks = temp_ranks;
  new_group->size = n;
  return 0;
}


async int MPI_Group_excl(MPI_Group group, int n, int *ranks, 
			 MPI_Group *new_group) {
  int i,temp_size,mem_index,mask;
  int *temp_members;
  int *temp_ranks;
  MPI_Group diff; 
  /* return 11: empty group, can't build subgroup */
  /* return 12: subgroup greater than main group  */
  
  if((!group.ranks)||(!group.members)) return 11;
  if(n > group.size) return 12;
  
  diff.size = n;
  diff.members = (int *) /**/shmalloc(MEMBER_ARR_SIZE * sizeof(int));
  diff.ranks   = (int *) /**/shmalloc(n * sizeof(int));
  
  for(i=0; i<MEMBER_ARR_SIZE; i++) diff.members[i] = 0;
  for(i=0; i<n; i++) {
    /* build new 'ranks'-array */
    mem_index = (group.ranks[ranks[i]])/(SYSTEM_WORD_LENGTH * sizeof(int));
    mask = 1 << (group.ranks[ranks[i]] %(SYSTEM_WORD_LENGTH * sizeof(int)));
    diff.members[mem_index] |= mask;
    diff.ranks[i] = group.ranks[ranks[i]];
  }
  MPI_Group_difference(group, diff, new_group);
  MPI_Group_free(&diff);
  return 0;
}

async int MPI_Group_size(MPI_Group group, int *size) {
  *size = group.size;
  return 0;
}


async int MPI_Group_rank(MPI_Group group, int *rank) {
  int i=0;
  *rank = -1;
  while(i < group.size) {
    if(group.ranks[i] == __PROC_NR__) {
      *rank = i;
      break;
    }
    i++;
  }
  return 0;
}

async int MPI_Group_compare(MPI_Group g1, MPI_Group g2, int *res) {
  /* since topologies are not (yet) supported, 
     compare won't EVER return 'congruent'     */
  
  int i,min;
  *res = MPI_IDENT;
  if((min = g1.size) != g2.size) {
    *res = MPI_UNEQUAL;
    return 0;
  }
  for(i=0; i<MEMBER_ARR_SIZE; i++) {
    if(g1.members[i] != g2.members[i]) {
      *res = MPI_UNEQUAL;
      return 0;
    }
  }
  for(i=0; i<g1.size; i++) {
    if(g1.ranks[i] != g2.ranks[i]) {
      *res = MPI_SIMILAR;
      return 0;
    }
  }
  *res = MPI_IDENT;
  return 0;
}


async int MPI_Group_free(MPI_Group *group) {
  if(group) {
    if(group->members) {
      shfree(group->members);
      group->members = NULL;
    }
    if(group->ranks) {
      shfree(group->ranks);
      group->ranks = NULL;
    }
    group->size = 0;
    return 0;
  }
  else  {
    return 1;
  }
}

async int MPI_Comm_group(MPI_Comm comm, MPI_Group *group) {
  MPI_Group_free(group);
  group->size = comm.group.size;
  group->members = (int *) shmalloc(MEMBER_ARR_SIZE * sizeof(int));
  memcpy(group->members, comm.group.members, MEMBER_ARR_SIZE * sizeof(int));
  group->ranks = (int *) shmalloc(group->size * sizeof(int));
  memcpy(group->ranks, comm.group.ranks, group->size * sizeof(int));
  return 0;
}

async int MPI_Comm_create(MPI_Comm comm, 
			  MPI_Group group, 
			  MPI_Comm *new_comm) {
  /* 
   * error values:
   * 11:     group isn't subset of comm.group 
   */
  
  int i;
  for(i=0; i<MEMBER_ARR_SIZE; i++)
    if((comm.group.members[i] | group.ranks[i]) 
       > comm.group.members[i]) 
      return 11;
  new_comm->group.size = group.size;
  new_comm->group.members= (int *) shmalloc(MEMBER_ARR_SIZE * sizeof(int));
  memcpy(new_comm->group.members, group.members,MEMBER_ARR_SIZE * sizeof(int));
  new_comm->group.ranks = (int *) shmalloc(group.size * sizeof(int));
  memcpy(new_comm->group.ranks, group.ranks,comm.group.size * sizeof(int));
  return 0;
}

async int MPI_Comm_free(MPI_Comm *comm) {
  MPI_Group_free(&(comm->group));
}


async int MPI_Comm_size(MPI_Comm comm, int *size) {
  MPI_Group group;
  Init_group(&group);
  MPI_Comm_group(comm, &group);
  MPI_Group_size(group, size);
  return 0;
}

/* not yet tested! */
async int MPI_Comm_rank(MPI_Comm comm, int *rank) {
  MPI_Group group;
  Init_group(&group);
  MPI_Comm_group(comm, &group);
  MPI_Group_rank(group, rank);
  return 0;
}

async int MPI_Comm_dup(MPI_Comm comm, MPI_Comm *new_comm) {
  return MPI_Comm_create(comm, comm.group, new_comm);
}

async int MPI_Comm_split(MPI_Comm comm, 
			 int color, int key, MPI_Comm *new_comm) {
  /* not yet implemented */
}

async int MPI_Init(int *argc, char ***argv) {
  int i,mask,ret = 0;
  
  MEMBER_ARR_SIZE = 1 + (__STARTED_PROCS__-1) / (SYSTEM_WORD_LENGTH * sizeof(int));
  if(mpadd(&whoisfirst,1) == 0) { 
    /* to make sure that only ONE P. does this */
    mailbox = (msgList **) shmalloc( __STARTED_PROCS__ * sizeof(msgList **)); 
    /* allocate mailbox array */
    if (!mailbox) ret = 1;
    globalMessageCounter = 0;
  }
  barrier;
  if(ret) return ret;
  /* initialize each proc's mailbox */ 
  mailbox[__PROC_NR__] =  	 
    (msgList *) shmalloc(__STARTED_PROCS__ * sizeof(msgList));
  if(! mailbox[__PROC_NR__]) return 1;
  for(i=0; i<__STARTED_PROCS__; i++) {
    /* allocate message list heads */
    mailbox[i][__PROC_NR__] = 
      (msgList) shmalloc (sizeof(struct msgList_struct));
    mailbox[i][__PROC_NR__]->lock = (simple_lock *) new_SimpleLock();
    mailbox[i][__PROC_NR__]->first = NULL;
    mailbox[i][__PROC_NR__]->last = NULL;     
  }  
  MPI_GROUP_EMPTY.size = 0;
  MPI_GROUP_EMPTY.members = (int *) /**/
    shmalloc(MEMBER_ARR_SIZE * sizeof(int));
  for(i=0; i<MEMBER_ARR_SIZE; i++) MPI_GROUP_EMPTY.members[i] = 0; 
  MPI_GROUP_EMPTY.ranks = NULL;
  
  MPI_COMM_WORLD.group.size = __STARTED_PROCS__;
  MPI_COMM_WORLD.group.members = 
    (int *) /**/shmalloc(MEMBER_ARR_SIZE * sizeof(int));
  set_all_bits(MPI_COMM_WORLD.group.members,__STARTED_PROCS__);
  MPI_COMM_WORLD.group.ranks = 
    (int *) /**/shmalloc(__STARTED_PROCS__ * sizeof(int));
  for(i=0; i < __STARTED_PROCS__; i++) 
    MPI_COMM_WORLD.group.ranks[i] = i;
  barrier;
  return ret;
}

async int MPI_Finalize(void) {
  int i;
  msg m, old_m;
  
  barrier;
  for(i=0; i < __STARTED_PROCS__; i++) {
    m = mailbox[i][__PROC_NR__]->first; 
    /* empty buffer lists */
    while(m) { 
      old_m = m;
      m = m->next;
      shfree(old_m->buffer);
      shfree(old_m);
    } 
    /* deallocate message list heads */
    shfree(mailbox[i][__PROC_NR__]);
  }
  /* deallocate each proc's mailbox */
  shfree(mailbox[__PROC_NR__]);
  barrier;
  if(__PROC_NR__ == 0) shfree(mailbox);
  return 0;
}

async int MPI_Sizeof(MPI_Datatype datatype) {
  return 1;
  /* Obviously only valid for SB-PRAM systems... */
}


async int mpi_sum(void *out,  void *in, 
		  int  *len, MPI_Datatype *type) {
  int i;
  int *ia, *ib;
  float *fa, *fb;
  if(*type == MPI_INT) {
    ia = (int *) out;
    ib = (int *) in;
    for(i=0; i < *len; i++) ia[i] += ib[i];
  }
  else if(*type == MPI_FLOAT) {
    fa = (float *) out;
    fb = (float *) in;
    for(i=0; i < *len; i++) fa[i] += fb[i];
  }
  else return 11; /* Op not provided for this type */
  return 0;
}

async int MPI_Reduce(void *sendbuf, void *recvbuf, int count,
 		     MPI_Datatype datatype, MPI_Op op, int root,
 		     MPI_Comm comm) {
  
  /* uses tag -1 for intra-communication */
  
  void *buf;
  int ret,i,size,myrank, softrank, depth, pred, succ;
  int *p;
  MPI_Group group;
  MPI_Status status;
  
  Init_group(&group);
  MPI_Comm_group(comm, &group);
  MPI_Group_size(group,&size);
  MPI_Group_rank(group, &myrank);
  
  softrank = myrank;
  depth = 0;
  buf = (void *) shmalloc(count * sizeof(datatype));
  prS("\n");
  while((softrank >= 0) && (size > 1)) {
    if(softrank % 2) {
      /* send message to predecessor (with even number) and quit */
      pred = myrank - (1<<depth);
      MPI_Send(sendbuf, count, datatype, pred,-1, comm);
      p = (int *) sendbuf;
      //pprintf("Sent %4d %4d %4d %4d %4d %4d %4d %4d\n", 
		//   p[0],p[1],p[2],p[3],p[4],p[5],p[6],p[7]);
      //pprintf("sent to %d, size = %d\n", pred, size);
      if(myrank == root) softrank = -3; else softrank = -1;
    }
    else {
      /* receive message from successor (with odd number) */
      if(softrank < size-1) {
	/* at least if there IS a successor... */
	succ = myrank + (1<<depth);
	MPI_Recv(buf, count, datatype, succ,
		 -1,comm, &status);
	p = (int *) sendbuf;
	if(! __PROC_NR__) {
	  pprintf("sendbuf: %4d %4d %4d %4d %4d %4d %4d %4d\n", 
		 p[0],p[1],p[2],p[3],p[4],p[5],p[6],p[7]);
	  p = (int *) buf;
	  pprintf("    buf: %4d %4d %4d %4d %4d %4d %4d %4d\n", 
		  p[0],p[1],p[2],p[3],p[4],p[5],p[6],p[7]);
	  //pprintf("received from %d, size = %d\n",
		    //succ, size);
	  mpi_sum(sendbuf, buf, &count, &datatype);
	  p = (int *) sendbuf;
	  pprintf("==============================================\n           sum: %4d %4d %4d %4d %4d %4d %4d %4d\n\n", 
		  p[0],p[1],p[2],p[3],p[4],p[5],p[6],p[7]);
	}
      }
      softrank = softrank / 2;
      size = size - (size / 2);
      depth++;
    }
  }
  if(softrank == -3) {
    MPI_Recv(recvbuf, count, datatype, group.ranks[0], -1, comm, &status);
  //pprintf("received from %d\n",group.ranks[softrank]);
  }
  else if(myrank == root) {
    MPI_Send(sendbuf, count, datatype, group.ranks[root], -1, comm);
    //pprintf("sent to %d\n",group.ranks[softrank-1]);
    //pprintf("Done!!!!!\n");
  }
  shfree(buf);
  return 0;
}


async void output(void) {
  int i,j;
  msg tmp;
  prS("========================\n");
  for(i=0; i<__STARTED_PROCS__; i++) {
    for(j=0; j<__STARTED_PROCS__; j++) {
      printf("\nproc[%d,%d]: ",i,j);
      tmp = mailbox[i][j]->first;
      while(tmp) {
	printf("->%p\t", (char *) tmp);
	tmp = tmp->next;
      }
    }
    prS("\n");
  }
  prS("========================\n\n\n\n");
}

async void print_group(MPI_Group group) {
  int i;
  int size;
  MPI_Group_size(group, &size);
  printf("\nsize   :%4d\n\nmembers:",size);
  if(group.members) for(i=0; i<MEMBER_ARR_SIZE;i++) 
    printf("%4d ", group.members[i]);
  else printf("(no members field)");
  prS("\n\nranks  :");
  if(group.ranks) for(i=0; i<group.size; i++) 
    printf("%2d ", group.ranks[i]);
  else printf("(no ranks field)");
  prS("\n");
}




