/* Gauss-Elimination in Fork95.   C.W.Kessler 02/96 */

#include <fork.h>
#include <io.h>
#include <math.h>
#include <assert.h>
#include <string.h>     // prototype for memcpy() - important!

/* The forall() macro cannot be nested.
 * We apply explicit flattening to a self-scheduling parallel loop
 * to exploit 2D parallelism.
 */


sync float **build_system( sh int *pn ) 
{
   /* build initial system of *pn equations in *pm variables.
    * RHS vector stored in column *pn of system matrix.
    */
  pr int i;
  pr int n;
  sh float **s;
  sh int p = groupsize();
  n = 3;    /*example values*/
  seq s = (float **) shmalloc( n * sizeof( float *));
  farm
    forall (i, 0, n, p)
      s[i] = (float *) shmalloc( (n+1) * sizeof( float )); 
  s[0][0] = 2.0;   s[0][1] =  -1.0;   s[0][2] = 3.0;  s[0][3]=6.0;
  s[1][0] =-3.0;   s[1][1] =   3.0;   s[1][2] =-2.0;  s[1][3]=4.0;
  s[2][0] = 1.0;   s[2][1] =   2.0;   s[2][2] = 1.0;  s[2][3]=0.0; 
  *pn = n;
  return s;
}


async void print_system( float **s, int n, int m )
{
  /* One processor prints the system of n equations
   * and m variables in sequential. */
  pr int i, j;
  printf("New %d x %d system:\n", n, m );
  printf("==================================================\n");
  for (i=0; i<n; i++) {
     printf("EQ%d: ", i);
     for (j=0; j<m; j++) 
        printf(" %5.2lf x%.2d", s[i][j], j );
     printf(" == %5.2lf\n", s[i][m] );
  }
  printf("==================================================\n");
}


typedef struct { float val; int loc; } pair;

sync void reduce( async sh void (*f)(void *, void *, void *),
                  sh void *s, sh void **d, sh int n, sh int elsize )
{
 sh int p = 0; 
 sh void **temp;
 int i;
 sh int t;
 $ = mpadd(&p,1); // assert consecutive numbering of $ in 0..p-1
 
 seq printf("reduce nach %p (%f) von %p size %d elsize %d\n", s, 
            *(float *)s, d, n, elsize);
 temp = (void **)shalloc( p * elsize );  // temporary array
 farm // initialize temp by the first p elements of d:
   if ($<n)
      memcpy( temp+$*elsize, d+$*elsize, elsize );

 seq {
   printf("Operandenvektor:\n");
   for (i=0; i<n; i++)
      printf(" %d: loc %d  val %f\n", i, ((pair *)temp)[i].loc, ((pair *)temp)[i].val );
 }
 // p partial reductions being computed:
 farm
   forall( i, p, n, p ) 
     f( temp + $*elsize, temp + $*elsize, d + i*elsize );

 // iterative computation upwards the tree:
 for (t=1; t<p; t = t<<1 )  // sequential loop over tree levels
   farm
     if ($+t < p) 
       f( temp+$*elsize, temp+$*elsize, temp + ($+t)*elsize );
 seq
   memcpy( s, temp, elsize );    // write result to *s
 shallfree();  // release temp
}



async void fabsmax( void *m, void *x, void *y )
{
 if (fabs(((pair*)x)->val) >= fabs(((pair*)y)->val)) {
    ((pair*)m)->val = ((pair *)x)->val;  ((pair*)m)->loc = ((pair*)x)->loc;
 }
 else {
    ((pair*)m)->val = ((pair *)y)->val;  ((pair*)m)->loc = ((pair*)y)->loc;
 }
}


sync void pivotize( sh float **a, sh int n, sh int r )
{
  sh float *hlp;
  sh int p = 0;
  sh pair piv;
  sh pair *col = (pair *)shalloc((n-r)*sizeof(pair));
  int i;
  $ = mpadd(&p, 1);
  for (i=$; i<n-r; i+=p) {
     col[i].loc = r+i;
     col[i].val = a[i+r][r];
  }
  reduce( fabsmax, &piv, (void **)col, n-r, sizeof(pair) );

  seq printf("Pivot row is %d, pivot is %f\n", piv.loc, piv.val);
  farm assert( piv.loc >= r );
  //farm assert( fabs(piv.val - a[piv.loc][r]) < 0.0001 );

  // permute rows:
  hlp = a[r];
  a[r] = a[piv.loc];
  a[piv.loc] = hlp;
}


sync void eliminate( sh float **a, sh int n )
{
  /* applies Gaussian elimination to system a[n][n+1].
   * Assume that matrix a is regular and diagonally dominant.
   * For a system of n equations in n variables,
   * up to n^2 processors could be used to update in parallel.
   * The solution is returned in the right hand side column a[*][n] 
   * of the modified system tableau.
   */
 int i, j, ij;
 sh int r;
 sh int p = 0;
 sh float factor;
 sh float *rowfactor = (float *)shalloc( n * sizeof(float));
 $ = mpadd( &p, 1 );   // renumber
 
 for (r=0; r<n; r++) {  /* do n elimination steps */ 

    if (r<n-1)  pivotize( a, n, r );

    factor = 1.0 / a[r][r];

    seq printf("Stage %d factor %lf p=%d\n", r, factor, p);
    /* my pivot element is a[r][r] because a is diag. dom. */

    /* normalize equation r: */
    farm
      forall ( j, r, n+1, p ) 
        a[r][j] = a[r][j] * factor;

    farm
      forall ( i, 0, n, p ) 
        rowfactor[i] = a[i][r];   // copying avoids data dependencies

    farm {
     forall2( i, j, ij, 0, n, r, n+1, p ) {
      //i = ij / (n-r+1);  // row index of element to be updated
      if (i!=r) {
        /* update row i: */
        //j = r + ij % (n-r+1);  // column index of element to be updated
        a[i][j] = a[i][j] - a[r][j] * rowfactor[i]; 
      }
     }
    }
    farm if ($==0) print_system( a, n, n );
 }  /* for r */
 return;
}


sh float **s;  /* system tableau */
sh int n;      /* n eq's in m var's */

void main( void )
{
  pr int starttime, stoptime;
  start {
    s = build_system( &n );
    farm if ($==0) print_system( s, n, n );
    farm starttime = getct();
    eliminate( s, n );
  }
  stoptime = getct();
  if ($==0) printf("\nTime: %d PRAM CPU cycles\n", stoptime - starttime);
  barrier;
}
